/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.math.BigInteger;

/**
 *
 * @author Michael Gilleland
 */

// Implementación de Michael Gilleland del algoritmo descrito por Kenneth H. Rosen en su libro
// Matemática discreta y sus aplicaciones.
// Ver 'http://www.merriampark.com/comb.htm'.

public class CombinationGenerator {

    private int n;                  // Número de elementos. Permanece fijo una vez ejecutado el constructor.
    private int m;                  // Número de tomas. Permanece fijo una vez ejecutado el constructor.
    private BigInteger total;       // Número total de combinaciones. Permanece fijo una vez ejecutado el constructor.
    private int[] v;                // Vector de combinaciones. Su tamaño permanece fijo una vez ejecutado el constructor.
    private BigInteger numLeft;     // Número de combinaciones todavía no generadas.

    // Constructor.
    public CombinationGenerator(int n, int m) {
        if (m > n || m < 0) throw new IllegalArgumentException ();
        if (n < 1) throw new IllegalArgumentException ();

        this.n = n;
        this.m = m;

        BigInteger nFact = getFactorial(n);
        BigInteger mFact = getFactorial(m);
        BigInteger nminusmFact = getFactorial(n - m);
        total = nFact.divide(mFact.multiply(nminusmFact));
        v = new int[m];

        reset();
    }

    // Repone el sistema de cálculo de combinaciones a la situación inicial.
    private void reset() {
        for (int i = 0; i < v.length; i++) {    // 'v.length' es igual a 'm'.
            v[i] = i;
        }
        numLeft = new BigInteger(total.toString());     // Quedan todas las combinaciones.
    }

    // Devuelve el número de combinaciones todavía no generadas.
    public BigInteger getNumLeft() {
        return numLeft;
    }

    // Devuelve un valor booleano que indica si quedan más combinaciones por generar.
    public boolean hasMore() {
        return (numLeft.compareTo(BigInteger.ZERO) == 1);
    }

    // Devuelve el número de combinaciones de n elementos tomados m a m.
    public BigInteger getTotal() {
        return total;
    }

    // Devuelve el factorial del número n.
    private static BigInteger getFactorial(int n) {
        BigInteger fact = BigInteger.ONE;
        for (int i = n; i > 1; i--) {
            fact = fact.multiply(new BigInteger(Integer.toString(i)));
        }
        return fact;
    }

    // Devuelve la siguiente combinación de n elementos tomados m a m.
    public int[] getNext() {
        if (numLeft.equals(total)) {
            numLeft = numLeft.subtract(BigInteger.ONE);
            return v;
        }

        int i = m - 1;
        while (v[i] == (n - m + i)) {
            i--;
        }
        
        v[i] = v[i] + 1;
        for (int j = (i + 1); j < m; j++) {
            v[j] = v[i] + j - i;
        }

        numLeft = numLeft.subtract(BigInteger.ONE);
        return v;
    }   
}