/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.*;
import javax.xml.parsers.ParserConfigurationException;
import matlabcontrol.MatlabConnectionException;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;
import matlabcontrol.MatlabProxyFactory;
import matlabcontrol.MatlabProxyFactoryOptions;
import org.xml.sax.SAXException;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
public class Control implements ActionListener {

    ModeloSistema modelo;
    Interfaz ventana;
    GestorFicheros gestorFicheros;
    GestorBBDD gestorBBDD;
    ConversorModelos moduloCM;
    GeneradorExperimentos moduloGE;
    GeneradorModelosSimulacion moduloGMS;
    Diagnosis moduloDX;

    //Método de creación
    public Control() {
        System.out.println("CONTROL class has been initialized");

        gestorFicheros = new GestorFicheros();
        gestorBBDD = new GestorBBDD();
        ventana = new Interfaz(this);
        ventana.validate();

        //Dimensiones de la ventana
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension frameSize = ventana.getSize();
        if (frameSize.height > screenSize.height) {
            frameSize.height = screenSize.height;
        }
        if (frameSize.width > screenSize.width) {
            frameSize.width = screenSize.width;
        }
        ventana.setLocation((screenSize.width - frameSize.width) / 2,
                (screenSize.height - frameSize.height) / 2);
        ventana.setVisible(true);

        modelo = new ModeloSistema();
        moduloCM = new ConversorModelos(modelo);
        moduloGE = new GeneradorExperimentos(modelo);
        moduloGMS = new GeneradorModelosSimulacion(modelo);
        moduloDX = new Diagnosis(modelo);
    }

    public static void main(String[] args) {
        // TODO code application logic here
        try {
            UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception exception) {
        }
        new Control();
    }

    //Lanza una ventana de error con el mensaje que le indiquemos
    public void muestraError(String err) {
        JOptionPane.showMessageDialog(ventana, err, "Error", JOptionPane.ERROR_MESSAGE);
    }

    //Crea una ventana en la que se pide confirmación al usuario sobre cualquier acción
    public int muestraConfirm(String confirmacion) {
        return JOptionPane.showConfirmDialog(ventana, confirmacion, "Warning", JOptionPane.YES_NO_OPTION);
    }

    //Administrador de eventos
    @Override
    public void actionPerformed(ActionEvent evento) {
            //eventos del menú "Archivo"
        if (evento.getActionCommand().startsWith("Salir")) {
            System.exit(1);
        } else if (evento.getActionCommand().startsWith("Entradas")) {
            System.out.println("'Inputs' File Generation");
            eventoGenerarEntrada();
        } else if (evento.getActionCommand().startsWith("Fallos")) {
            System.out.println("'Faults' File Genetarion");
            eventoGenerarFallos();

            //eventos del menú "Ventana"
        } else if (evento.getActionCommand().startsWith("Abrir panel CM")) {
            ventana.panelPestanas.setSelectedIndex(0);
        } else if (evento.getActionCommand().startsWith("Abrir panel GE")) {
            ventana.panelPestanas.setSelectedIndex(1);
        } else if (evento.getActionCommand().startsWith("Abrir panel GMS")) {
            ventana.panelPestanas.setSelectedIndex(2);
        } else if (evento.getActionCommand().startsWith("Abrir panel DX")) {
            ventana.panelPestanas.setSelectedIndex(3);

            //eventos del menú "Ayuda"
        } else if (evento.getActionCommand().startsWith("Contenidos")) {
            VentanaAyuda v = new VentanaAyuda();
            v.setVisible(true);
        } else if (evento.getActionCommand().startsWith("Acerca de")) {
            VentanaAcercaDe v = new VentanaAcercaDe();
            v.setVisible(true);

            //eventos del módulo "Conversor de modelos"
        } else if (evento.getActionCommand().startsWith("Cargar Modelo CM")) {
            System.out.println("Load model");
            eventoCargarModeloSistemaCM();
        } else if (evento.getActionCommand().startsWith("convertir")) {
            if (!moduloCM.ficherosCargados) {
                muestraError("No system loaded");
            } else {
                System.out.println("Convert model into I-O notation");
                eventoConvertirModelo();
            }
        } else if (evento.getActionCommand().startsWith("GuardarModES")) {
            if (!moduloCM.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloCM.modeloConvertido) {
                muestraError("You must convert the model first");
            } else {
                System.out.println("Save model in I-O notation");
                eventoGuardarModelo();
            }

            //eventos del módulo "Generador de modelos de simulación"
        } else if (evento.getActionCommand().startsWith("Cargar Modelo GMS")) {
            System.out.println("Load model");
            eventoCargarModeloSistemaGMS();
        } else if (evento.getActionCommand().startsWith("GenerarSim")) {
            if (!moduloGMS.ficherosCargados) {
                muestraError("No system loaded");
            } else {
                System.out.println("Generate Simulation Models");
                modelo.listaModelosSimulacion = eventoGenerarModeloSimulacion();
            }
        } else if (evento.getActionCommand().startsWith("GuardarModSim")) {
            if (!moduloGMS.ficherosCargados) {
                System.out.println("*** 'save' option failure in SMG module");
                muestraError("No system loaded");
            } else if (!moduloGMS.modelosGenerados) {
                muestraError("You must generate the Simulation Models first");
            } else {
                System.out.println("Save Simulation Models");
                eventoGuardarModeloSimulacion();
            }

            //eventos del módulo "Generador de experimentos"
        } else if (evento.getActionCommand().startsWith("Cargar Modelo GE")) {
            System.out.println("Load model in EG panel");
            eventoCargarModeloSistemaGE();
        } else if (evento.getActionCommand().startsWith("GenerarExp")) {
            if (!moduloGE.ficherosCargados) {
                muestraError("No system loaded");
            } else {
                System.out.println("Generate Experiment");
                eventoGenerarExperimento();
            }
        } else if (evento.getActionCommand().startsWith("VerGrafico")) {
            if (!moduloGE.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloGE.experimentoGenerado) {
                muestraError("You must generate an experiment first");
            } else {
                System.out.println("Display Graph");
                eventoVerGrafico();
            }
        } else if (evento.getActionCommand().startsWith("GuardarExp")) {
            if (!moduloGE.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloGE.experimentoGenerado) {
                muestraError("You must generate an experiment first");
            } else {
                System.out.println("Save experiment results");
                eventoGuardarExperimento();
            }

            //eventos del módulo "Diagnosis"
        } else if (evento.getActionCommand().startsWith("Cargar Modelo DX")) {
            System.out.println("Load model in DX panel");
            eventoCargarModeloSistemaDX();
        } else if (evento.getActionCommand().startsWith("Calcular Residuos")) {
            if (!moduloDX.ficherosCargados) {
                muestraError("No system loaded");
            } else {
                System.out.println("Residual calculation");
                eventoCalcularResiduos();
            }
        } else if (evento.getActionCommand().startsWith("GraficoDX")) {
            if (!moduloDX.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloDX.residuosCalculados) {
                muestraError("You must calculate the residuals first");
            } else {
                System.out.println("Display graph");
                eventoVerGraficoDX();
            }
        } else if (evento.getActionCommand().startsWith("Calcular Candidatos")) {
            if (!moduloDX.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloDX.residuosCalculados) {
                muestraError("You must calculate the residuals first");
            } else {
                System.out.println("Candidate calculation");
                eventoCalcularCandidatos();
            }
        } else if (evento.getActionCommand().startsWith("GuardarDX")) {
            if (!moduloDX.ficherosCargados) {
                muestraError("No system loaded");
            } else if (!moduloDX.residuosCalculados) {
                muestraError("You must calculate the residuals and the fault candidates first");
            } else if (!moduloDX.candidatosCalculados) {
                muestraError("You must calculate the fault candidates first");
            } else {
                System.out.println("Save diagnosis results");
                eventoGuardarDiagnosis();
            }
        }
    }

    //Este evento se activa cuando el usuario selecciona la opción
    //"Generar fichero de entrada" en el menú. Para cada variable de entrada
    //se generan tantos valores como instantes de simulación
    private void eventoGenerarEntrada() {
        modelo = new ModeloSistema();
        gestorFicheros.cargarVarEntradaSistema(modelo);
        generarEntradas();
    }

    //Función principal de la generación de valores para las variables de entrada.
    //Puede lanzarse desde el evento anterior o dentro del proceso de carga de
    //datos del módulo GE
    public String generarEntradas() {
        if (modelo.modeloCargadoGeneradorEntradas()) {
            ArrayList<double[]> valores = new ArrayList<double[]>();
            boolean seleccionValida = true;
            GeneradorFicheroEntradas gfe = new GeneradorFicheroEntradas(ventana);
            if (gfe.valido) {
                int tsim = gfe.Tsim;

                //para cada variable de entrada se crea un GeneradorFicheroEntradas, el cual se encarga
                //de presentar los formularios adecuados y recoger las entradas del usuario
                Iterator i = modelo.variablesEntrada.iterator();
                while (i.hasNext() && seleccionValida) {
                    String variable = (String) i.next();
                    gfe = new GeneradorFicheroEntradas(ventana, variable, tsim);

                    if (gfe.valido) {
                        valores.add(gfe.generarValoresEntrada());   //si el formulario se rellenó correctamente, se generan los valores
                    } else {
                        seleccionValida = false;
                    }
                }
            } else {
                seleccionValida = false;
            }

            //sólo se guardan los resultados si el proceso de generación de valores se completó
            //satisfactoriamente para TODAS las variables
            if (seleccionValida) {
                FiltroFicheros filtro = new FiltroFicheros("txt");
                JFileChooser fc = new JFileChooser();
                fc.setDialogTitle("Save Inputs file");
                fc.setFileFilter(filtro);
                int respuesta = fc.showSaveDialog(fc);
                if (respuesta == JFileChooser.APPROVE_OPTION) {
                    boolean escribir = false;
                    if (fc.getSelectedFile().exists()) {
                        if (muestraConfirm("The file already exists, ¿do you want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                            escribir = true;
                        }
                    } else {
                        escribir = true;
                    }
                    if (escribir) {
                        String nomFichero = fc.getSelectedFile().getAbsolutePath();
                        if (!nomFichero.endsWith(".txt")) {
                            nomFichero = nomFichero + ".txt";
                        }
                        gestorFicheros.guardarFicheroEntradas(modelo.variablesEntrada, valores, nomFichero);
                        return nomFichero;
                    }
                }
            }
        } else {
            muestraError("Input data has not been loaded correctly. Please, try again.");
        }
        return null;
    }

    //Este evento se activa cuando el usuario selecciona la opción
    //"Generar fichero de fallos paramétricos" en el menú. 
    private void eventoGenerarFallos() {
        modelo = new ModeloSistema();
        gestorFicheros.cargarParametros(modelo);
        generarFallos();
    }

    //Función principal de la generación de fallos paramétricos.
    //Puede lanzarse desde el evento anterior o dentro
    //del proceso de carga de datos del módulo GE
    public String generarFallos() {
        if (modelo.modeloCargadoGeneradorFallos()) {
            DialogoParametros dp = new DialogoParametros(ventana, modelo);  //el usuario selecciona los parámetros que causarán fallo; se almacenan sus índices
            ArrayList<Integer> indices = new ArrayList<Integer>();
            if (dp.valido) {
                for (int i = 0; i < dp.jcb.length; i++) {
                    if (dp.jcb[i] != null && dp.jcb[i].isSelected()) {
                        indices.add(i);
                        System.out.println("*** Parameter selected: " + dp.jcb[i].getText());
                    }
                }
            }

            //con los índices de los parámetros de fallo, se llama a GeneradorFicheroFallos
            //para que construya y almacene los fallos correspondientes
            GeneradorFicheroFallos gff = new GeneradorFicheroFallos(ventana, indices, modelo);
            ArrayList<Fallo> fallos = gff.generarFallos();

            //se almacenan los fallos en un fichero de texto
            if (fallos != null && !fallos.isEmpty()) {
                FiltroFicheros filtro = new FiltroFicheros("txt");
                JFileChooser fc = new JFileChooser();
                fc.setDialogTitle("Save failures file");
                fc.setFileFilter(filtro);
                int respuesta = fc.showSaveDialog(fc);
                if (respuesta == JFileChooser.APPROVE_OPTION) {
                    boolean escribir = false;
                    if (fc.getSelectedFile().exists()) {
                        if (muestraConfirm("The file already exists, ¿do you want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                            escribir = true;
                        }
                    } else {
                        escribir = true;
                    }
                    if (escribir) {
                        String nomFichero = fc.getSelectedFile().getAbsolutePath();
                        if (!nomFichero.endsWith(".txt")) {
                            nomFichero = nomFichero + ".txt";
                        }
                        gestorFicheros.guardarFicheroFallos(fallos, nomFichero);
                        return nomFichero;
                    }
                }
            }
        } else {
            muestraError("Input data has not been loaded correctly. Please, try again.");
        }
        return null;
    }

    //Evento producido al pulsar sobre el botón "Cargar Modelo"
    //del panel del conversor de modelos
    private void eventoCargarModeloSistemaCM() {
        modelo = new ModeloSistema();
        moduloCM = new ConversorModelos(modelo);
        
        Object selectedValue = panelSeleccionCarga();
        if (selectedValue != null && selectedValue.toString().compareTo("Path File") == 0) {
            gestorFicheros.cargarDatosCM(modelo);   //se carga todo lo que necesita el módulo
        } else if (selectedValue != null && selectedValue.toString().compareTo("Manual Selection") == 0) {
            gestorFicheros.cargarEcuacionesSistema(modelo);     //se cargan las EDOs
            gestorFicheros.cargarVarEntradaSistema(modelo);     //se cargan las variables del sistema    
        }
        
        if (modelo.modeloCargadoModuloCM()) {
            moduloCM.setFicherosCargados();
            ventana.imprimeModeloPanelCM(modelo);
        } else {
            muestraError("Input data has not been loaded correctly. Please, try again.");
        }
    }

    //Evento producido al pulsar sobre el botón "Convertir modelo"
    //del panel del conversor de modelos
    private void eventoConvertirModelo() {
        ArrayList<String> r = moduloCM.convertirModelo();   //se almacenan en una lista las líneas del modelo convertido por moduloCM
        ventana.imprimeModESPanelCM(r);
        modelo.setModeloES(r);
        moduloCM.setModeloConvertido();
    }

    //Evento producido al pulsar sobre el botón "Guardar"
    //del panel del conversor de modelos
    private void eventoGuardarModelo() {
        FiltroFicheros filtro = new FiltroFicheros("txt");
        JFileChooser fc = new JFileChooser();
        fc.setFileFilter(filtro);
        int respuesta = fc.showSaveDialog(fc);
        if (respuesta == JFileChooser.APPROVE_OPTION) {
            boolean escribir = false;
            if (fc.getSelectedFile().exists()) {
                if (muestraConfirm("The file already exists, ¿do you want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                    escribir = true;
                }
            } else {
                escribir = true;
            }
            if (escribir) {
                String nomFichero = fc.getSelectedFile().getAbsolutePath();
                if (!nomFichero.endsWith(".txt")) {
                    nomFichero = nomFichero + ".txt";
                }
                gestorFicheros.guardarModeloConvertido(modelo, nomFichero);   //Se guardan los resultados en fichero
            }
        }
    }

    //Evento producido al pulsar sobre el botón "Cargar Modelo"
    //del panel de generador de experimentos
    private void eventoCargarModeloSistemaGE() {
        modelo = new ModeloSistema();
        moduloGE = new GeneradorExperimentos(modelo);
        Object selectedValue = panelSeleccionCarga();
        if (selectedValue != null && selectedValue.toString().compareTo("Path File") == 0) {
            gestorFicheros.cargarDatosGE(modelo, moduloGE);     //se carga todo lo que necesita el módulo
        } else if (selectedValue != null && selectedValue.toString().compareTo("Manual Selection") == 0) {
            gestorFicheros.cargarEcuacionesSistema(modelo);     //se cargan las EDOs
            gestorFicheros.cargarVarEntradaSistema(modelo);     //se cargan las variables del sistema

            // A continuación se carga o se introduce la configuración del experimento
            selectedValue = panelSeleccionConfiguracion();
            if (selectedValue != null && selectedValue.toString().compareTo("Load from Existing File") == 0) {
                gestorFicheros.cargarConfiguracionExperimento(moduloGE);    //carga la configuración del experimento
            } else if (selectedValue != null && selectedValue.toString().compareTo("Create new File interactively") == 0) {
                //el usuario introduce la configuración mediante un formulario. Queda guardada directamente
                //en moduloGE, pero luego se permite también almacenarla en un fichero
                FormularioConfiguracion fc = new FormularioConfiguracion(ventana, true, moduloGE);
                if (moduloGE.configExpCargada()) {
                    int seleccion = JOptionPane.showOptionDialog(
                            ventana,
                            "¿Do you wanto to save the experiment configuration?", //Mensaje
                            "Select an option", // Título
                            JOptionPane.YES_NO_CANCEL_OPTION,
                            JOptionPane.QUESTION_MESSAGE,
                            null,
                            new Object[]{"Yes", "No"},
                            "Yes");
                    if (seleccion != -1) {
                        if ((seleccion + 1) == 1) {
                            guardarConfiguracion();
                        } else {
                            // PRESIONO NO
                        }
                    }
                }
            }

            gestorFicheros.cargarParametros(modelo);    //se cargan los parámetros del sistema

            // A continuación se cargan o se generan los valores de las variables de entrada.
            selectedValue = panelSeleccionCargaEntradas();
            if (selectedValue != null && selectedValue.toString().compareTo("Load from Existing File") == 0) {
                gestorFicheros.cargarValoresEntrada(moduloGE);      //carga los valores de las variables de entrada
            } else if (selectedValue != null && selectedValue.toString().compareTo("Create new File interactively") == 0) {
                String ruta = generarEntradas();    //el usuario crea interactivamente un fichero con los valores
                if (ruta != null) {
                    gestorFicheros.cargarValoresEntrada(moduloGE, ruta);    //se carga el fichero creado
                }
            }

            gestorFicheros.cargarCondicionesIniciales(moduloGE);    //carga de las condiciones iniciales

            // A continuación se cargan o se introducen los valores de los fallos paramétricos.
            selectedValue = panelSeleccionCargaFallos();
            if (selectedValue != null && selectedValue.toString().compareTo("Load from Existing File") == 0) {
                gestorFicheros.cargarFallos(moduloGE);      //se cargan los fallos paramétricos
            } else if (selectedValue != null && selectedValue.toString().compareTo("Create new File interactively") == 0) {
                String ruta = generarFallos();  //el usuario crea interactivamente un fichero con los fallos
                if (ruta != null) {
                    gestorFicheros.cargarFallos(moduloGE, ruta);    //se carga el fichero creado
                }
            }
        }
        
        if (modelo.modeloCargadoModuloGE() && moduloGE.modeloCargadoModuloGE()) {
            moduloGE.setFicherosCargados();
            ventana.imprimeModeloPanelGE(modelo);
        } else {
            muestraError("Unable to load input data correctly. Please, try again.");
        }
    }

    //Almacena en un fichero la configuración del experimento
    //que el usuario introdujo de forma interactiva
    public void guardarConfiguracion() {
        FiltroFicheros filtro = new FiltroFicheros("config");
        JFileChooser fc = new JFileChooser();
        fc.setDialogTitle("Save experiment configuration");
        fc.setFileFilter(filtro);
        int respuesta = fc.showSaveDialog(fc);
        if (respuesta == JFileChooser.APPROVE_OPTION) {
            boolean escribir = false;
            if (fc.getSelectedFile().exists()) {
                if (muestraConfirm("The file already exists, ¿do you want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                    escribir = true;
                }
            } else {
                escribir = true;
            }
            if (escribir) {
                String nomFichero = fc.getSelectedFile().getAbsolutePath();
                if (!nomFichero.endsWith(".config")) {
                    nomFichero = nomFichero + ".config";
                }
                gestorFicheros.guardarConfiguracionExperimento(moduloGE, nomFichero);
            }
        }
    }

    //Evento producido al pulsar sobre el botón "Generar Experimento"
    //del panel de generador de experimentos
    private void eventoGenerarExperimento() {
        modelo.prepararSistema();       // Ordena las ecuaciones y genera los modelos observacional y transicional.
        ArrayList<String> variablesComputo = moduloGE.generarFuncionInstante();     //se construye el fichero .m ejecutable por matlab
        Object[] o = variablesComputo.toArray();
        String[] v = Arrays.copyOf(o, o.length, String[].class);
        try {
            ventana.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            moduloGE.ejecutarExperimento(v);    //se ejecuta el experimento, usando el fichero creado anteriormente
            ventana.setCursor(Cursor.getDefaultCursor());
            moduloGE.setExperimentoGenerado();
            ventana.imprimeTablaExperimento(moduloGE.modeloTabla);
        } catch (MatlabConnectionException ex) {
            Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        } catch (MatlabInvocationException ex) {
            Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    //Evento producido al pulsar el botón "Ver Gráfico"
    //del panel de generador de experimentos
    private void eventoVerGrafico() {
        DialogoVariablesGrafico dvg = new DialogoVariablesGrafico(ventana, moduloGE);   //el usuario selecciona las variables que desea visualizar

        ArrayList<String> variables = new ArrayList<String>();
        if (dvg.valido) {
            System.out.println("***** Variables selected *****");
            for (int i = 0; i < dvg.jcb.length; i++) {
                if (dvg.jcb[i] != null && dvg.jcb[i].isSelected()) {
                    variables.add(dvg.jcb[i].getText());
                    System.out.println(dvg.jcb[i].getText());
                }
            }

            //si se seleccionaron varias variables, se construye un gráfico general; si sólo fue una,
            //se genera un gráfico un poco más personalizado
            if (variables.size() > 1) {
                moduloGE.mostrarGrafico(variables);
            } else {
                System.out.println(variables.size());
                moduloGE.mostrarGrafico(variables.get(0));
            }
        }
    }

    //Evento producido al pulsar sobre el botón "Guardar"
    //del panel de generador de experimentos
    private void eventoGuardarExperimento() {
        Object selectedValue = panelSeleccionGuardarExperimento();
        if (selectedValue != null && selectedValue.toString().compareTo("Save in Text File") == 0) {
            FiltroFicheros filtro = new FiltroFicheros("txt");
            JFileChooser fc = new JFileChooser();
            fc.setFileFilter(filtro);
            int respuesta = fc.showSaveDialog(fc);
            if (respuesta == JFileChooser.APPROVE_OPTION) {
                boolean escribir = false;
                if (fc.getSelectedFile().exists()) {
                    if (muestraConfirm("The file already exists, ¿do you want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                        escribir = true;
                    }
                } else {
                    escribir = true;
                }
                if (escribir) {
                    String nomFichero = fc.getSelectedFile().getAbsolutePath();
                    if (!nomFichero.endsWith(".txt")) {
                        nomFichero = nomFichero + ".txt";
                    }
                    gestorFicheros.guardarResultadoExperimento(moduloGE, nomFichero);   //Se guardan los resultados en fichero
                }
            }
        } else if (selectedValue != null && selectedValue.toString().compareTo("Save in Database") == 0){
            FormularioAccesoBBDD fa = new FormularioAccesoBBDD(ventana, true);
            if (fa.valido) {
                FormularioGuardarExpBBDD fg = new FormularioGuardarExpBBDD(ventana, true);
                if (fg.valido) {
                    ventana.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
                    gestorBBDD.guardarExperimento(moduloGE, ventana, fg.getDescripcion(), fg.getNombreSistema(), fa.getNombreBBDD(), fa.getUsuario(), fa.getPassword());
                    ventana.setCursor(Cursor.getDefaultCursor());
                }
            }
        }
    }

    //Evento producido al pulsar sobre el botón "Cargar Modelo"
    //del panel de generador de modelos de simulación
    private void eventoCargarModeloSistemaGMS() {
        modelo = new ModeloSistema();
        moduloGMS = new GeneradorModelosSimulacion(modelo);
        Object selectedValue = panelSeleccionCarga();
        if (selectedValue != null && selectedValue.toString().compareTo("Path File") == 0) {
            gestorFicheros.cargarDatosGMS(modelo);      //se carga todo lo que necesita el módulo
        } else if (selectedValue != null && selectedValue.toString().compareTo("Manual Selection") == 0) {
            gestorFicheros.cargarEcuacionesSistema(modelo);     //se cargan las EDOs
            gestorFicheros.cargarVarEntradaSistema(modelo);     //se cargan las variables de sistema
            gestorFicheros.cargarModeloCPC(modelo);     //se cargan los resultados del cálculo de PCs
        }
        
        if (modelo.modeloCargadoModuloGMS()) {
            moduloGMS.setFicherosCargados();
            ventana.imprimeModeloPanelGMS(modelo);
            ventana.imprimeCPCPanelGMS(modelo);
        } else {
            muestraError("Unable to load input data correctly. Please, try again.");
        }
    }

    //Evento producido al pulsar sobre el botón "Generar Simulación"
    //del panel de generador de modelos de simulación
    private ArrayList<String> eventoGenerarModeloSimulacion() {
        modelo.prepararSistema();
        //pasamos los MEMs a una lista
        ArrayList<MEM> modelosSimulacion = new ArrayList<MEM>();
        for (int i = 0; i < modelo.modelos.size(); i++) {
            modelosSimulacion.add(modelo.modelos.get(i).clonar());
        }
        ArrayList<String> lista = moduloGMS.generarModelosSimulacion(modelosSimulacion);    //se generan los modelos de simulación (uno por cada MEM)
        ventana.imprimeModSimPanelGMS(lista);
        moduloGMS.setModelosGenerados();
        return lista;
    }

    //Evento producido al pulsar sobre el botón "Guardar"
    //del panel de generador de modelos de simulación
    private void eventoGuardarModeloSimulacion() {
        gestorFicheros.guardarModelosSimulacion(modelo);
    }

    private void eventoCargarModeloSistemaDX() {
        modelo = new ModeloSistema();
        moduloDX = new Diagnosis(modelo);
        GeneradorExperimentos moduloAuxGE = new GeneradorExperimentos(modelo);
        Object selectedValue = panelSeleccionCarga();
        if (selectedValue != null && selectedValue.toString().compareTo("Path File") == 0) {
            gestorFicheros.cargarDatosDX(modelo, moduloAuxGE);      //se carga todo lo que necesita el módulo
        } else if (selectedValue != null && selectedValue.toString().compareTo("Manual Selection") == 0) {
            gestorFicheros.cargarVarEntradaSistema(modelo);     //se cargan las variables del sistema
            gestorFicheros.cargarModSim(modelo);    //se cargan los modelos de simulación
            gestorFicheros.cargarConfiguracionExperimento(moduloAuxGE);     //se carga la configuración de un experimento
            gestorFicheros.cargarParametros(modelo);    //se cargan los parámetros del sistema
            gestorFicheros.cargarCondicionesIniciales(moduloAuxGE);     //se cargan las condiciones iniciales
            gestorFicheros.cargarModeloCPC(modelo);     //se cargan los resultados del cálculo de PCs ****NOTA: ESTA INFORMACIÓN YA NO SE USA
            gestorFicheros.cargarFallos(moduloAuxGE);
            selectedValue = panelSeleccionCargarExperimento();
            if (selectedValue != null && selectedValue.toString().compareTo("Load from Text File") == 0) {
                gestorFicheros.cargarExperimento(moduloAuxGE);      //se cargan los resultados de un experimento desde fichero
            } else {
                FormularioAccesoBBDD f = new FormularioAccesoBBDD(ventana, true);
                if (f.valido) {
                    gestorBBDD.cargarExperimento(moduloAuxGE, ventana, f.getNombreBBDD(), f.getUsuario(), f.getPassword());
                }
            }
        }
        
        if (modelo.modeloCargadoModuloDX() && moduloAuxGE.modeloCargadoModuloDX()) {
            moduloDX.setFicherosCargados();
            moduloDX.setModuloAuxiliarGE(moduloAuxGE);
            ventana.imprimeModeloPanelDX(modelo);
            ventana.imprimeTablaExpDX(moduloAuxGE.modeloTabla);
        }
        
        ArrayList <String> listaCEMs = new ArrayList <String>();
        for (int i = 0; i < modelo.listaModSim.size(); i++) {
            String nombreModelo = modelo.listaModSim.get(i).nombreModelo;
            Pattern patron = Pattern.compile("Modelo \\d+\\((.+)\\)");
            Matcher matcher = patron.matcher(nombreModelo);
            matcher.find();
            String cem = matcher.group(1);
            if (modelo.listaModSim.get(i).nodoDiscrepancia.size() > 0) {
                String nodo = modelo.listaModSim.get(i).nodoDiscrepancia.get(0).nombre.replace("_2", "").replace("mes", "") + "(" + nombreModelo + ")";
                if (!listaCEMs.contains(cem)) {
                    listaCEMs.add(cem);
                    moduloDX.nodosDiscrepancia.add(nodo);;
                }
            }
        }
    }

    private void eventoCalcularResiduos() {
        //el usuario elige de qué modo se calculan los residuos
        Object selectedValue = panelSeleccionModoResiduos();
        if (selectedValue == null) {
            return;
        }else if (selectedValue.toString().compareTo("Differences") == 0) {
            moduloDX.setModoCalculoResiduos("Diferencia");
        } else if (selectedValue != null && selectedValue.toString().compareTo("Differences Absolute Values") == 0) {
            moduloDX.setModoCalculoResiduos("Valor absoluto");
        } else if (selectedValue != null && selectedValue.toString().compareTo("Euclidean Distances") == 0) {
            moduloDX.setModoCalculoResiduos("Distancia euclídea");
        }

        //para cada modelo de simulación, se genera un fichero .m para Matlab y se ejecuta
        ventana.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        try {
            MatlabProxyFactoryOptions options = new MatlabProxyFactoryOptions.Builder().setHidden(true).build();
            MatlabProxyFactory factory = new MatlabProxyFactory(options);
            MatlabProxy proxy = factory.getProxy();
            for (int i = 0; i < modelo.listaModSim.size(); i++) {
                //System.out.println("Modelo " + i);
                moduloDX.generarFuncionSimulacion(modelo.listaModSim.get(i), modelo);
                System.out.flush();
                moduloDX.ejecutarSimulacion(proxy, modelo.listaModSim.get(i), modelo);
                proxy.eval("clear");
            }
            proxy.exit();
            proxy.disconnect();
        } catch (MatlabConnectionException ex) {
                Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        } catch (MatlabInvocationException ex) {
                Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        ventana.setCursor(Cursor.getDefaultCursor());

        moduloDX.construirTablaResiduos(modelo.listaModSim);
        ventana.imprimeTablaResiduos(moduloDX.modeloTabla);
        moduloDX.setResiduosCalculados();
    }
    
    private void eventoVerGraficoDX() {
        //el usuario selecciona la variable que desea visualizar
        ArrayList <ModeloSimulacion> listaModelos = modelo.listaModSim;
        ArrayList <Variable> nodosDisc = new ArrayList <Variable>();
        ArrayList <String> modelos = new ArrayList <String>();
        for (int i = 0; i < listaModelos.size(); i++) {
            nodosDisc.add(listaModelos.get(i).nodoDiscrepancia.get(0));
            modelos.add(listaModelos.get(i).nombreModelo);
        }
        Object[] posiblesValores = new Object[nodosDisc.size()];
        for (int i = 0; i < posiblesValores.length; i++) {
            posiblesValores[i] = nodosDisc.get(i).nombre + " (" +modelos.get(i) + ")";
        }
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select discrepance node to display its graph", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        
        if (selectedValue != null) {
            System.out.println(selectedValue);
            moduloDX.verGrafico(selectedValue.toString());
        }
    }

    public void eventoCalcularCandidatos() {
        // Primero, el usuario debe decidir qué tipo de candidatos quiere obtener
        // - Ecuaciones
        // - Componentes
        Object tipoCand = panelSeleccionTipoCand();
        if (tipoCand == null) {
            return;
        } else if (tipoCand.toString().startsWith("Components") || tipoCand.toString().startsWith("Equations")) {
            moduloDX.tipoCandidatos = tipoCand.toString();
        }
        
        try {
            gestorFicheros.cargarSistemaXML(ventana, moduloDX);     //se cargan los resultados del cálculo de PCs en formato XML
        } catch (ParserConfigurationException ex) {
            Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        } catch (SAXException ex) {
            Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(Control.class.getName()).log(Level.SEVERE, null, ex);
        }

        // El usuario selecciona el método que define cómo se activan los PCs. Hay 3 opciones:
        // - Usar un valor de umbral global (definido en la configuración del experimento).
        // - Permitir al usuario introducir un umbral distinto para cada PC.
        // - Realizar un Z-Test.
        
        double[] umbrales = new double[moduloDX.nodosDiscrepancia.size()];
        boolean variosUmbrales = false;
        Object selectedValue = panelSeleccionUmbral();
        
        if (selectedValue == null) {
            return;
        } else if (selectedValue.toString().startsWith("Insert threshold")) {   //si se seleccionó usar varios umbrales, se recogen los datos con un formulario
            variosUmbrales = true;
            DialogoUmbrales du = new DialogoUmbrales(ventana, moduloDX.nodosDiscrepancia);
            if (du.valido) {
                for (int i = 0; i < du.jtf.length; i++) {
                    if (du.jtf[i] != null) {
                        umbrales[i] = Double.parseDouble(du.jtf[i].getText());
                        System.out.println("*** The threshold for " + du.jl[i].getText() + " is " + umbrales[i]);
                    }
                }
            }
        } else if (selectedValue.toString().startsWith("Use the global value")) {
            // No necesitamos hacer nada. El algoritmo tomará en su momento
            // el valor definido para los umbrales.
        } else {
            // Z-test
        }

        // Se buscan instante a instante los PCs que se activan. Cuando se produce
        // una activación, se calculan los candidatos para los PCs activos en ese
        // momento. También se almacena información del instante en que se produjo
        // la activación y las CEMs que intervinieron.
        // Si el usuario seleccionó realizar un Z-Test, se pasa la responsabilidad
        // de todo este proceso a un módulo de Lógica de Decisión.
        ArrayList<ArrayList<String[]>> candidatos;
        ArrayList<String> ModelosActivados;
        ArrayList<Double> instantesActivacion;
        int numDiscrepancias = moduloDX.nodosDiscrepancia.size();

        if (selectedValue.toString().startsWith("Run Z-test")) { //se hará un Z-Test
            LogicaDecision ld = new LogicaDecision(moduloDX);
            new FormularioZtest(ventana, true, ld);     //se presenta un formulario en el que el usuario introduce los parámetros del Z-Test
            ld.calcularCandidatos(modelo);    //se realiza todo el proceso de activación de PCs y cálculo de candidatos
            candidatos = ld.candidatos;
            ModelosActivados = ld.modelosActivados;
            instantesActivacion = ld.instantesActivacion;
        } else {    //se trabaja con uno o más umbrales
            candidatos = new ArrayList<ArrayList<String[]>>();
            ModelosActivados = new ArrayList<String>();
            instantesActivacion = new ArrayList<Double>();

            boolean[] banderas = new boolean[numDiscrepancias];     //cada bandera indica si la CEM correspondiente se ha activado
            for (int i = 0; i < banderas.length; i++) {
                banderas[i] = false;
            }

            String nodoDiscrepancia;
            double residuo;
            double umbralActual = moduloDX.moduloAuxGE.umbral;  //en principio se usa el umbral global
            
            int numInstantes = moduloDX.modeloTabla.getRowCount();
            for (int i = 0; i < numInstantes; i++) {
                for (int j = 0; j < numDiscrepancias; j++) {
                    if (!banderas[j]) {     //si un modelo se activa, permanece así hasta el final y ya no se mira si tiene más activaciones
                        String nombreModelo = modelo.listaModSim.get(j).nombreModelo;
                        nodoDiscrepancia = moduloDX.nodosDiscrepancia.get(j);
                        residuo = moduloDX.getValorResiduoTablaRes(i, nombreModelo);
                        
                        Pattern patron = Pattern.compile("Modelo (\\d+)\\(Cadena (\\d+)\\)");
                        Matcher matcher = patron.matcher(nombreModelo);
                        matcher.find();
                        int cadena = Integer.parseInt(matcher.group(2));
                        int index = 0;
                        
                        String cadenaAux;
                        for (int k = 0; k < moduloDX.cemNumber; k++) {
                            cadenaAux = moduloDX.cem[k].name;
                            if (cadenaAux.contains(" " + cadena)) {
                                index = k;
                                break;
                            }
                        }
                        if (index >= moduloDX.cemNumber) {
                            System.out.println("An error was found while calculating the candidates.");
                            return;
                        }
                                
                        if (variosUmbrales) {
                            umbralActual = umbrales[j];     //si se seleccionó usar un umbral para cada PC, se pasa a usar el umbral correspondiente
                        }
                        if (Math.abs(residuo) > Math.abs(umbralActual)) {   //se produce una activación
                            System.out.println("\nInstant: " + moduloDX.getValorTablaRes(i, "Time") +"; Residual: " + residuo + "; Threshold: " + umbralActual);
                            banderas[j] = true;
                            moduloDX.activarCEM(index);
                            ModelosActivados.add(nombreModelo);
                            instantesActivacion.add(moduloDX.getValorTablaRes(i, "Time"));
                            moduloDX.calculaCandidatos(modelo);   //se calculan los candidatos según el estado de activaciones en el instante actual
                            candidatos.add((ArrayList<String[]>) moduloDX.cand.clone());
                            System.out.println("*** Node " + nodoDiscrepancia + " was activated at instant " + moduloDX.getValorTablaRes(i, "Time"));
                        }
                    }
                }
            }
        }

        ArrayList<Diagnosis.CEM> CEMsSistema = new ArrayList<Diagnosis.CEM>();
        for (int j = 0; j < moduloDX.cemNumber; j++) {
            CEMsSistema.add(moduloDX.getCEM(j));
        }

        ventana.imprimeCandidatos(candidatos, CEMsSistema, ModelosActivados, instantesActivacion, moduloDX.nodosDiscrepancia);
        moduloDX.setCandidatosCalculados();
    }

    public void eventoGuardarDiagnosis() {
        FiltroFicheros filtro = new FiltroFicheros("txt");
        JFileChooser fc = new JFileChooser();
        fc.setFileFilter(filtro);
        int respuesta = fc.showSaveDialog(fc);
        if (respuesta == JFileChooser.APPROVE_OPTION) {
            boolean escribir = false;
            if (fc.getSelectedFile().exists()) {
                if (muestraConfirm("The file already exists, ¿do yo want to overwrite it?") == 0) { //Se ha aceptado sobreescribir
                    escribir = true;
                }
            } else {
                escribir = true;
            }
            if (escribir) {
                String nomFichero = fc.getSelectedFile().getAbsolutePath();
                if (!nomFichero.endsWith(".txt")) {
                    nomFichero = nomFichero + ".txt";
                }
                gestorFicheros.guardarResultadoDiagnosis(ventana, nomFichero);
            }
        }
    }

    //Lanza un diálogo para que el usuario seleccione la forma de cargar
    //los distintos ficheros de entrada: manualmente, o a través de un fichero
    //que contenga las rutas los mismos
    public Object panelSeleccionCarga() {
        Object[] posiblesValores = {"Path File", "Manual Selection"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to load the input data", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    //Lanza un diálogo para que el usuario seleccione la forma de cargar
    //los valores de las variables de entrada: a través de un fichero ya
    //existente, o crearlo de forma interactiva
    public Object panelSeleccionCargaEntradas() {
        Object[] posiblesValores = {"Load from Existing File", "Create new File interactively"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to load the input variables values", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    //Lanza un diálogo para que el usuario seleccione la forma de cargar
    //los fallos paramétricos: a través de un fichero ya
    //existente, o crearlo de forma interactiva
    public Object panelSeleccionCargaFallos() {
        Object[] posiblesValores = {"Load from Existing File", "Create new File interactively"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to load parametric faults", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    //Lanza un diálogo para que el usuario seleccione la forma de cargar
    //los parámetros de configuración del experimento: a través de un fichero ya
    //existente, o introducirlos de forma interactiva
    public Object panelSeleccionConfiguracion() {
        Object[] posiblesValores = {"Load from Existing File", "Introduce interactively"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to load experiment configuration parameters", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    public Object panelSeleccionModoResiduos() {
        Object[] posiblesValores = {"Differences Absolute Values", "Differences", "Euclidean Distances"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to calculate the residuals", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    public Object panelSeleccionUmbral() {
        Object[] posiblesValores = {"Run Z-test", "Use the global value defined in the configuration file",
            "Insert threshold value for each Possible Conflict"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way the threshold for each PC will be defined", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }
    
    public Object panelSeleccionTipoCand() {
        Object[] posiblesValores = {"Equations", "Components"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the type of candidates that you want to search for", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    public Object panelSeleccionGuardarExperimento() {
        Object[] posiblesValores = {"Save in Text File", "Save in Database"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to save the experiment results", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }

    public Object panelSeleccionCargarExperimento() {
        Object[] posiblesValores = {"Load from Text File", "Extract from Database"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the way you want to load the experiment results", "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }
}