/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;
import javax.swing.table.DefaultTableModel;
import javolution.util.FastBitSet;
import matlabcontrol.MatlabConnectionException;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;
import org.jfree.chart.ChartFrame;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.AxisLocation;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.CombinedDomainXYPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.StandardXYItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
public class Diagnosis {

    Boolean ficherosCargados;
    Boolean residuosCalculados;
    Boolean candidatosCalculados;
    String modoCalculoResiduos;     //puede ser "Diferencia", "Valor absoluto" o "Distancia euclídea"
    ModeloSistema modelo;
    GeneradorExperimentos moduloAuxGE;
    String rutaFuncionSimulacion;
    DefaultTableModel modeloTabla;
    ArrayList<String> nodosDiscrepancia = new ArrayList<String>(); // Conjunto de todos los nodos discrepancia de los modelos cargados.
    // **** ATRIBUTOS NECESARIOS PARA EL CÁLCULO DE CANDIDATOS ****
    // Variables globales con toda la información del sistema.
    // Para almacenar la definición del sistema, las referencias y las
    // variables auxiliares para los cálculos.
    Document docXML;    // Mantiene el documento XML cargado para el cálculo de candidatos
    String tipoCandidatos;
    String fileDesc;
    int cemNumber;
    CEM[] cem;  // Contiene las diferentes cadenas con un PC asociado que hemos cargado.
    ArrayList<String> comp; // Para los distintos componentes.
    ArrayList<String[]> cand = new ArrayList(1024); // Para los candidatos.
    long elapsedTime; // Para el tiempo de cálculo de los candidatos.
    
    // Etiquetas, argumentos y constantes utilizados en el fichero de definición XML.
    final String XML_RESULTADOS = "RESULTADOS";     // Elemento 'RESULTADOS'. Es la raíz del documento XML.
    final String XML_CCEM = "CCEM";                 // Elemento 'CCEM', dependiente de 'RESULTADOS'. Es el conjunto de CEMs.
    final String XML_SIST = "sistema";              // Elemento 'sistema', que contiene a los componentes.
    final String XML_SYSTEM_NAME = "nom_stma";      // Atributo 'nom_stma' del elemento 'CCEM'. Es el nombre del sistema.
    final String XML_CEM = "CADENA";                // Elemento 'CADENA'. Es la definición del CEM.
    final String XML_CEM_NAME = "id_cadena";        // Atributo 'id_cadena' del elemento 'CADENA'. Es el nombre del CEM.
    final String XML_CEM_IS_PC = "si_PC";           // Atributo 'si_PC' del elemento 'CADENA'. Indica si es un posible conflicto.
    final String XML_CEM_EQUATIONS = "ecuaciones";  // Atributo 'ecuaciones'. Es la lista de ecuaciones que forman la Cadena.
    final String XML_COMPONENT = "componente";      // Elementos 'componente'
    final String XML_COMP_NAME = "nombre_com";      // Atributo 'nombre_com' de los elementos componente.
    final String XML_MEM = "MEM";                   // Elemento 'MEM' bajo el elemento 'CADENA'.
    final String XML_EQ = "ecuacion";
    final String XML_ID = "id";                     // Atributo 'id' de los elementos MEM y ecuación.
    final String XML_USED = "usado";                // Atributo 'usado' del elemento 'MEM'.
    final String XML_NAME = "nombre";               // Atributo 'nombre' de los elementos 'interpretacion', 'cabeza' y 'cola'.
    final String XML_YES = "si";                    // Valor 'si' de los atributos 'si_PC' (elemento 'CADENA'), 'usado' (elemento
                                                    // 'MEM') y 'observada' (elementos 'cabeza' y 'cola').
    //final String XML_OBSERVED = "observada";        // Atributo 'observada' de los elementos 'cabeza' y 'cola'
    //final String XML_INTERPR = "interpretaciones";  // Atributo 'interpretaciones' bajo el elemento 'MEM'.
    //final String XML_INTERPR_HEAD = "cabeza";       // Elmento 'cabeza' bajo el elemento 'interpretacion'.
    //final String XML_INTERPR_TAIL = "cola";         // Elmento 'cola' bajo el elemento 'interpretacion'.
   
    // Definiciones para la comprobación y la evaluación de las puertas lógicas.
    final int LG_AND = 0;
    final int LG_OR = 1;
    final int LG_XOR = 2;
    final int LG_NOT = 3;
    final int LG_NAND = 4;
    final int LG_NOR = 5;
    final int LG_XNOR = 6;
    // Las funciones tienen que ir en el mismo orden que las constantes LG_*
    final String[] logicGateName = {
        "AND", "OR", "XOR", "NOT", "NAND", "NOR", "XNOR"
    };
    ArrayList<String> compX;    // Variable global auxiliar (para los distintos componentes).

    public Diagnosis(ModeloSistema m) {
        modelo = m;
        ficherosCargados = false;
        residuosCalculados = false;
        candidatosCalculados = false;
    }

    public void setModoCalculoResiduos(String modo) {
        this.modoCalculoResiduos = modo;
    }

    public void setModuloAuxiliarGE(GeneradorExperimentos ge) {
        this.moduloAuxGE = ge;
    }

    public void setFicherosCargados() {
        ficherosCargados = true;
    }

    public void setResiduosCalculados() {
        residuosCalculados = true;
    }

    public void setCandidatosCalculados() {
        candidatosCalculados = true;
    }

    public void generarFuncionSimulacion(ModeloSimulacion modSim, ModeloSistema modelo) {
        this.rutaFuncionSimulacion = (new File(System.getProperty("user.dir"))).getParentFile().getAbsolutePath() + "\\FuncionInstante";  //Usar en la versión entregable
        //this.rutaFuncionSimulacion = System.getProperty("user.dir") + "\\FuncionInstante";  //Usar dentro de NetBeans o en la versión instalable
        String numFuncion = modSim.nombreModelo.replace("Modelo ", "").split("\\(")[0];
        String nomFichero = rutaFuncionSimulacion + "\\" + "simulation" + numFuncion + ".m";
        File fichero = new File(rutaFuncionSimulacion);
        fichero.setExecutable(true, false);
        fichero.setReadable(true, false);
        fichero.setWritable(true, false);
        String linea;

        try {
            BufferedWriter bw = new BufferedWriter(new FileWriter(nomFichero));
            linea = "%Simulation Model: " + modSim.nombreModelo;
            bw.write(linea);
            bw.newLine();
            linea = "function xNew = simulation(AntStateVar,Ent,DeltaT,Param,DiscNode,MesVar)";
            bw.write(linea);
            bw.newLine();
            bw.newLine();
            bw.write("% System inputs:");
            bw.newLine();

            if (!modSim.variablesEntrada.isEmpty()) {
                for (int i = 0; i < modSim.variablesEntrada.size(); i++) {
                    linea = modSim.variablesEntrada.get(i) + " = Ent(" + (i+1) + ") ;";
                    bw.write(linea);
                    bw.newLine();
                }
            } else {
                bw.write("% Not needed");
                bw.newLine();
            }
            
            bw.newLine();
            bw.write("% Previous instant predictions for the state variables:");
            bw.newLine();

            Iterator i = modSim.variablesEstado.iterator();
            int c = 1;
            while (i.hasNext()) {
                String var = (String) i.next();
                linea = var + "_ant = AntStateVar(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            
            bw.newLine();
            bw.write("% System Parameters:");
            bw.newLine();

            i = modelo.parametros.iterator();
            c = 1;
            while (i.hasNext()) {
                Parametro p = (Parametro) i.next();
                linea = p.nombre + " = Param(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            
            bw.newLine();
            bw.write("% Discrepancy nodes: ");
            bw.newLine();
            c = 1;
            for (int j = 0; j < modSim.nodoDiscrepancia.size(); j++) {
                String var = modSim.nodoDiscrepancia.get(j).nombre;
                linea = "%" + var + " = DiscNode(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            
            bw.newLine();
            bw.write("% Experiment variables measured: ");
            bw.newLine();
            i = modSim.variablesSalida.iterator();
            c = 1;
            while (i.hasNext()) {
                String var = (String) i.next();
                linea = var + " = MesVar(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            
            bw.newLine();
            bw.write("% Model equations: ");
            bw.newLine();
                    
            for (int l = 0; l < modSim.listaEcuaciones.size(); l++) {
                String ecuacion = modSim.listaEcuaciones.get(l);
                linea = ecuacion;
                if (!linea.contains(";")) {
                    linea = linea + " ;";
                }
                bw.write(linea);
                bw.newLine();
            }
            
            bw.newLine();
            bw.write("% Outputs:");
            bw.newLine();
            bw.write("%  - Values to update state vars");
            bw.newLine();
            bw.write("%  - Discrepancy nodes");
            bw.newLine();
            
            i = modSim.variablesEstado.iterator();
            c = 1;
            while (i.hasNext()) {
                String var = (String) i.next();
                linea = "xNew(" + c + ") = " + var + ";";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            for (int l=0;l<modSim.nodoDiscrepancia.size();l++) {
                String v = (String) modSim.nodoDiscrepancia.get(l).nombre;
                linea = "xNew(" + c + ") = " + v + ";";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            bw.newLine();
            linea = "end";
            bw.write(linea);
            bw.flush();
            bw.close();   
        } catch (IOException io) {
            io.printStackTrace();
        }
    }

    // Ejecuta repetidamente la función "simulation" tras crear una conexión con Matlab.
    // Los resultados de cada simulación se almacenan en la variable  nodoDiscrepancia,
    // propia de cada Modelo de Simulación.
    public void ejecutarSimulacion(MatlabProxy proxy, ModeloSimulacion modSim, ModeloSistema modelo) throws MatlabConnectionException, MatlabInvocationException {
        proxy.eval("addpath('" + this.rutaFuncionSimulacion + "')");
        String numFuncion = modSim.nombreModelo.replace("Modelo ", "").split("\\(")[0];

        double DeltaT = moduloAuxGE.deltaT;
        double instanteActual = roundFourDecimals(moduloAuxGE.getValorCondicionInicial("Time", 0));

        int numEntradas = modSim.variablesEntrada.size();
        double[] Ent = new double[numEntradas];
        for (int i = 0; i < numEntradas; i++) {
            Ent[i] = getValorTablaExp(0, modSim.variablesEntrada.get(i));
        }
        
        int numVarEstado = modSim.variablesEstado.size();
        double[] AntStateVar = new double[numVarEstado];
        for (int i = 0; i < numVarEstado; i++) {
            String varEstado = modSim.variablesEstado.get(i);
            AntStateVar[i] = moduloAuxGE.getValorCondicionInicial(varEstado, 0);
        }
        
        int numParametros = modelo.parametros.size();
        double[] Param = new double[numParametros];
        for (int i = 0; i < numParametros; i++) {
            Param[i] = modelo.parametros.get(i).valor;
        }
        
        int numDisc = modSim.nodoDiscrepancia.size();
        System.out.flush();
        double[] nodoDisc = new double[numDisc];
        for (int i = 0; i < numDisc; i++) {
            nodoDisc[i] = 0;
        }
        modSim.nodoDiscrepancia.get(0).addValor(0);   // Usada para compensar el retraso de var_ant
        
        int numVarSalida = modSim.variablesSalida.size();
        double[] mesVar = new double[numVarSalida];
        for (int i = 0; i < numVarSalida; i++) {
            String varSalida = modSim.variablesSalida.get(i);
            mesVar[i] = getValorTablaExp(0, varSalida);
        }

        // Se realiza una ejecución de simulation.m con las condiciones iniciales
        Object result = proxy.returningFeval("simulation" + numFuncion, 1, AntStateVar, Ent, DeltaT, Param, nodoDisc, mesVar)[0];
        
        double[] x = (double[]) result;        
        
        instanteActual = roundFourDecimals(instanteActual + DeltaT);
        int entradaActual = 1;

        // Se simulan el resto de instantes
        while (instanteActual <= moduloAuxGE.Tsim) {
            for (int i = 0; i < numEntradas; i++) {
                Ent[i] = getValorTablaExp(entradaActual, modSim.variablesEntrada.get(i));
            }
            for (int j = 0; j < numVarEstado; j++) {
                AntStateVar [j] = x[j];
            }
            for (int k = 0; k < numDisc; k++) {
                nodoDisc[k] = x[numVarEstado + k];
                modSim.nodoDiscrepancia.get(k).addValor(nodoDisc[k]);
            }
            for (int j = 0; j < numVarSalida; j++) {
                mesVar[j] = getValorTablaExp(entradaActual, modSim.variablesSalida.get(j));
            }
            
            result = proxy.returningFeval("simulation" + numFuncion, 1, AntStateVar, Ent, DeltaT, Param, nodoDisc, mesVar)[0];
            x = (double[]) result;
            instanteActual = roundFourDecimals(instanteActual + DeltaT);
            entradaActual = entradaActual + 1;
        }
        
        // Añadimos un valor al segundo nodo discrepancia para compensar el
        // cálculo del valor siguiente en lugar del actual de la variable de estado
        if (modSim.nodoDiscrepancia.size() > 1) {
            modSim.nodoDiscrepancia.get(1).addValor(roundFourDecimals(x[numVarEstado + 1]));
        }
    }

    public void construirTablaResiduos(ArrayList<ModeloSimulacion> modelosEjecutables) {
        
        ModeloSimulacion modSim;
        ArrayList<Variable> nodoDisc = new ArrayList <Variable>();
        
        //Creación de la tabla
        modeloTabla = new DefaultTableModel();
        modeloTabla.addColumn("Time");
        for (int i = 0; i < modelosEjecutables.size(); i++) {
            modSim = modelosEjecutables.get(i);
            nodoDisc = modSim.nodoDiscrepancia;
            int numDisc = nodoDisc.size();
            if (numDisc == 2) {
                for (int j = 0;j < nodoDisc.size(); j++) {
                    modeloTabla.addColumn(nodoDisc.get(j).nombre + " (" + modSim.nombreModelo + ")");
                }
            } else if (numDisc == 1) {
                modeloTabla.addColumn(nodoDisc.get(0).nombre + " (" + modSim.nombreModelo + ")");
                modeloTabla.addColumn(nodoDisc.get(0).nombre + " (Experiment)");
            }
            modeloTabla.addColumn("Residual (" + nodoDisc.get(0).nombre + "(" + modSim.nombreModelo +")");
        }
        
        // Metemos en la tabla todos los valores de la simulación
        int numInstantes = (nodoDisc.get(0).valores.size()-1);
        for (int i = 0; i <= numInstantes; i++) {
            Object[] fila = new Object[modeloTabla.getColumnCount()];
            int n = 0;
            fila[n] = moduloAuxGE.modeloTabla.getValueAt(i, 0);
            n++;
            
            double valorSim;
            double valorExp;
            for (int j = 0; j < modelosEjecutables.size(); j++) {
                modSim = modelosEjecutables.get(j);
                nodoDisc = modSim.nodoDiscrepancia;
                int numDisc = nodoDisc.size();
                if (numDisc == 2) {
                    valorSim = nodoDisc.get(0).valores.get(i);
                    valorExp = nodoDisc.get(1).valores.get(i);  
                } else if (numDisc == 1) {
                    valorSim = nodoDisc.get(0).valores.get(i);
                    valorExp = getValorTablaExp(i, nodoDisc.get(0).nombre);
                } else {
                    System.out.println("Error while printing the Simulation results. Discrepancy node not found");
                    return;
                }
                fila[n] = roundFourDecimals(valorSim);
                fila[n+1] = roundFourDecimals(valorExp);
                if (modoCalculoResiduos.compareTo("Valor absoluto") == 0) {
                    fila[n+2] = roundFourDecimals(Math.abs(valorExp - valorSim));
                } else if (modoCalculoResiduos.compareTo("Diferencia") == 0) {
                    fila[n+2] = roundFourDecimals((valorExp - valorSim));
                } else if (modoCalculoResiduos.compareTo("Distancia euclídea") == 0) {
                    fila[n+2] = roundFourDecimals(Math.sqrt(Math.pow((valorExp - valorSim), 2)));
                }
                n = n + 3;
            }
            modeloTabla.addRow(fila);
        }
    }

    public void verGrafico(String variable) {
        String [] partes = variable.split(" \\(Modelo");
        String varAux = partes [0];
        String nombreModelo = "Modelo" + partes [1].replace("))",")");
        
        XYSeries series1 = new XYSeries(variable);
        double instante;
        for (int i = 0; i < modeloTabla.getRowCount(); i++) {
            instante = new Double(modeloTabla.getValueAt(i, 0).toString());
            series1.add(instante, getValorTablaRes(i, variable));
        }
        
        // Para buscar los valores del nodo a mostrar, debemos renombrarlo
        // para que el nombre coincida con el almacenado en la tabla de residuos
        ArrayList <ModeloSimulacion> listaMods = modelo.listaModSim;
        Iterator n = listaMods.iterator();
        int numDisc = 0;
        while (n.hasNext()) {
            ModeloSimulacion modAux = (ModeloSimulacion) n.next();
            if (nombreModelo.equals(modAux.nombreModelo)) {
                numDisc = modAux.nodoDiscrepancia.size();
            }
        }
        if (numDisc > 1) {
            varAux = partes[0] + "_2 (" + nombreModelo + ")";
        } else {
            varAux = varAux + " (Experiment)";
        }
        XYSeries series2 = new XYSeries(varAux);
        for (int i = 0; i < modeloTabla.getRowCount(); i++) {
            instante = new Double(modeloTabla.getValueAt(i, 0).toString());
            series2.add(instante, getValorTablaRes(i, varAux));
        }

        XYSeriesCollection collection = new XYSeriesCollection();
        collection.addSeries(series1);
        collection.addSeries(series2);
        XYDataset data1 = collection;
        final XYItemRenderer renderer1 = new StandardXYItemRenderer();
        final NumberAxis rangeAxis1 = new NumberAxis("Values");
        final XYPlot subplot1 = new XYPlot(data1, null, rangeAxis1, renderer1);
        subplot1.setRangeAxisLocation(AxisLocation.BOTTOM_OR_LEFT);

        varAux = "Residual (" + varAux.split(" \\(Experiment\\)")[0] + "(" + nombreModelo + ")";
        XYSeries series3 = new XYSeries(varAux);
        for (int i = 0; i < modeloTabla.getRowCount(); i++) {
            instante = new Double(modeloTabla.getValueAt(i, 0).toString());
            series3.add(instante, getValorResiduoTablaRes(i, nombreModelo));
        }

        XYDataset data2 = new XYSeriesCollection(series3);
        final XYItemRenderer renderer2 = new StandardXYItemRenderer();
        final NumberAxis rangeAxis2 = new NumberAxis("Residuals");
        rangeAxis2.setAutoRangeIncludesZero(false);
        final XYPlot subplot2 = new XYPlot(data2, null, rangeAxis2, renderer2);
        subplot2.setRangeAxisLocation(AxisLocation.TOP_OR_LEFT);

        CombinedDomainXYPlot plot = new CombinedDomainXYPlot(new NumberAxis("Time"));
        plot.add(subplot1, 2);
        plot.add(subplot2, 1);
        plot.setOrientation(PlotOrientation.VERTICAL);
        JFreeChart chart = new JFreeChart("Chart for variable " + variable,
                JFreeChart.DEFAULT_TITLE_FONT, plot, true);

        ChartFrame frame = new ChartFrame("Chart", chart);
        frame.pack();
        frame.setVisible(true);
    }

    public void calculaCandidatos(ModeloSistema modelo) {
        ArrayList<String> compAux = comp;
        cand.clear();
        ArrayList<FastBitSet> cemMap = new ArrayList(1024);

        for (int i = 0; i < cemNumber; i++) {
            CEM cemItem = cem[i];
            if (cemItem.considerFlag) {
                FastBitSet cemMapItem = new FastBitSet(compAux.size());
                cemMapItem.clear();
                for (int j = 0; j < cemItem.candRef.size(); j++) {
                    cemMapItem.set(cemItem.candRef.get(j));
                }
                cemMap.add(cemMapItem);
            }
        }

        long startTime = System.currentTimeMillis(); // Se registra el momento de comienzo del algoritmo del 'hitting set'.
        // Se implementa el algoritmo de obtención del 'hitting set'.
        int[] cv = null;
        int[] dv;
        FastBitSet mapAux = new FastBitSet(compAux.size());
        ArrayList<FastBitSet> cemMapAux = new ArrayList(1024);
        int[] n;
        int[][] index;

        //System.out.println("INIT");
        for (int i = cemMap.size(); i > 0; i--) {
            int[] ccv = null;
            if (i < cemMap.size()) {
                ccv = new int[cemMap.size() - i];
            }
            boolean[] calCompVec = new boolean[cemMap.size()];
            CombinationGenerator cgen = new CombinationGenerator(cemMap.size(), i);
            //System.out.println("Combinations of " + i + " PCs have been generated");
            while (cgen.hasMore()) {
                cemMapAux.clear();
                if (i == 1) {
                    for (int j = 0; j < cemMap.size(); j++) {
                        cemMapAux.add(cemMap.get(j));
                    }
                } else {
                    
                    // Se obtiene el conjunto intersección.
                    cv = cgen.getNext();
                    //System.out.println("  PCs combination " + Arrays.toString(cv));
                    mapAux.set(0, compAux.size()); // Pone todos los bits a uno.
                    boolean doContinue = false;
                    for (int j = 0; j < cv.length; j++) {
                        mapAux.and(cemMap.get(cv[j]));
                        if (mapAux.cardinality() == 0) { // 'cardinality' devuelve el número de bits puestos a 1.
                            doContinue = true;
                            break;
                        }
                    }
                    if (doContinue) {
                        //System.out.println("    Empty intersection; system proceeds to the next combination");
                        continue;
                    } // Si el conjunto de intersección está vacío, se salta al siguiente.
                    
                    // Se dispone un vector con el conjunto intersección y todos los conjuntos que
                    // no han participado en la misma.
                    cemMapAux.add(mapAux);
                    if (i < cemMap.size()) {
                        // Se obtienen los índices de los conjuntos que no han participado en la intersección.
                        for (int j = 0; j < cemMap.size(); j++) {
                            calCompVec[j] = true;
                        }
                        for (int j = 0; j < cv.length; j++) {
                            calCompVec[cv[j]] = false;
                        }
                        for (int j = 0, k = 0; j < cemMap.size(); j++) {
                            if (calCompVec[j]) {
                                ccv[k++] = j;
                            }
                        }
                        for (int j = 0; j < ccv.length; j++) {
                            FastBitSet cemMapItem = cemMap.get(ccv[j]);
                            if (cemMapItem.cardinality() > 0) {
                                cemMapAux.add(cemMapItem);
                            }
                        }
                    }
                }
                
                // Se obtienen las distintas combinaciones entre los componentes de la
                // intersección y los componentes de los conjuntos que no han participado
                // en la intersección.
                n = new int[cemMapAux.size()];
                index = new int[cemMapAux.size()][];
                for (int j = 0; j < cemMapAux.size(); j++) {
                    FastBitSet cemMapItem = cemMapAux.get(j);
                    int numComp = cemMapItem.cardinality();
                    n[j] = numComp;
                    int[] indexItem = new int[numComp];
                    for (int k = 0, bitIndex = 0; k < numComp; k++) {
                        bitIndex = cemMapItem.nextSetBit(bitIndex);
                        indexItem[k] = bitIndex++;
                    }
                    index[j] = indexItem;
                }

                //System.out.println("    n = " + Arrays.toString(n));
                //System.out.println("    index = " + Arrays.deepToString(index));
                //System.out.println("    Layout generation from n");

                DispositionGenerator dgen = new DispositionGenerator(n);
                while (dgen.hasMore()) {
                    dv = dgen.getNext();
                    //System.out.println("      Equation layout " + Arrays.toString(dv));
                    String[] candItem = new String[dv.length];
                    for (int j = 0; j < dv.length; j++) {
                        //System.out.println("        Added " + compAux.get(index[j][dv[j]]) + " to candidate");
                        candItem[j] = compAux.get(index[j][dv[j]]);
                    }
                    cand.add(candItem);
                }
                if (i == 1) {
                    break; // En este caso, no hay que repetir el proceso para cada combinación.
                } else {
                    //System.out.println("        Equation presented in the intersection was deleted");
                    // Se eliminan los compentes de la intersección.
                    // Se eliminan los conjuntos que se quedan sin elementos.
                    // Se modifica el contador 'i' de acuerdo al nuevo número de conjuntos con elementos.
                    mapAux.flip(0, compAux.size());
                    for (int j = 0; j < cv.length; j++) {
                        FastBitSet cemMapItem = cemMap.get(cv[j]);
                        cemMapItem.and(mapAux);
                    }
                }
            }
            // Se eliminan los conjuntos que no tienen bits activados.
            if (i > 1) {
                for (int j = (cemMap.size() - 1); j >= 0; j--) {
                    if (cemMap.get(j).cardinality() == 0) {
                        //System.out.println("Empty set deleted, and decremented i counter");
                        cemMap.remove(j);
                        i--;
                    }
                }
            }
        }
        //System.out.println("END");
        elapsedTime = System.currentTimeMillis() - startTime;
    }

    public double getValorTablaExp(int fila, String columna) {
        for (int i = 0; i < moduloAuxGE.modeloTabla.getColumnCount(); i++) {
            if (columna.compareTo(moduloAuxGE.modeloTabla.getColumnName(i)) == 0) {
                return new Double(moduloAuxGE.modeloTabla.getValueAt(fila, i).toString());
            }
        }
        System.out.println("Column " + columna + " not found in the experiment table.");
        return -1;
    }

    public double getValorTablaRes(int fila, String columna) {
        for (int i = 0; i < modeloTabla.getColumnCount(); i++) {
            if (columna.compareTo(modeloTabla.getColumnName(i)) == 0) {
                return new Double(modeloTabla.getValueAt(fila, i).toString());
            }
        }
        System.out.println("Column " + columna + " not found in the residuals table.");
        return -1;
    }
    
    public double getValorResiduoTablaRes(int fila, String columna) {
        for (int i = 0; i < modeloTabla.getColumnCount(); i++) {
            if (modeloTabla.getColumnName(i).startsWith("Residual (") && (modeloTabla.getColumnName(i).contains(columna))) {
                return new Double(modeloTabla.getValueAt(fila, i).toString());
            }
        }
        System.out.println("Residual " + columna + " not found.");
        return -1;
    }

    //Redondea valores a cinco decimales.
    public double roundFourDecimals(double d) {
        double result = d * 10000;
        result = Math.round(result);
        result = result / 10000;
        return result;
    }
    
    // ****** CLASES PARA EL CÁLCULO DE CANDIDATOS *****
    // Definición de estructuras de datos.
    public static class CEM {
        String name;                                                // Nombre del CEM.
        ArrayList candName = new ArrayList();                       // Nombres de los componentes.
        ArrayList <Integer> candRef = new ArrayList <Integer>();    // Referencias a los componentes en el vector general de componentes.
        boolean considerFlag;                                       // Indicador de CEM en el que se han dado discrepancias y debe ser utilizado para la búsqueda de candidatos.
    }

    // **** FUNCIONES AUXILIARES PARA EL CÁLCULO DE CANDIDATOS ****
    public int cargaComp(String compName) {
        int i;
        for (i = 0; i < compX.size(); i++) {
            if (compX.get(i).equals(compName)) {
                break;
            }
        }
        if (i >= compX.size()) {
            compX.add(compName);
        }
        return (i);
    }
    
    // Recorre el documento XML generado por la herramienta de Posibles Conflictos, 
    // y almacena los elementos que va encontrando, creando una lista de posibles
    // candidatos, asociada a cada CEM.
    public void cargaSistema() /*throws Exception*/ {
        compX = new ArrayList();
        try {
            String fileDescX = docXML.getDocumentElement().getAttribute(XML_SYSTEM_NAME);   // Se obtiene el nombre del sistema.
            
            // Se calcula el número de CEMs y se reserva memoria.
            NodeList listaCEMs = docXML.getElementsByTagName(XML_CEM);
            int cemNumberX = 0;
            for (int i = 0; i < listaCEMs.getLength(); i++) {
                Node nodo_i = listaCEMs.item(i);
                if (nodo_i.getNodeType() == Node.ELEMENT_NODE) {
                    Element elementoCEM = (Element) nodo_i;
                    if (elementoCEM.getAttribute(XML_CEM_IS_PC).equalsIgnoreCase(XML_YES)) {
                        cemNumberX++;
                    }
                }
            }
            CEM[] cemX = new CEM[cemNumberX];

            // Para cada CEM pertinente se buscan los componentes/ecuaciones asociados,
            // que serán los posibles candidatos a fallo entre los que hay que buscar.
            int n = 0;
            listaCEMs = docXML.getElementsByTagName(XML_CEM);
            for (int i = 0; i < listaCEMs.getLength(); i++) {
                Node nodo_i = listaCEMs.item(i);
                if (nodo_i.getNodeType() == Node.ELEMENT_NODE) {
                    Element elementoCEM = (Element) nodo_i;
                    if (elementoCEM.getAttribute(XML_CEM_IS_PC).equalsIgnoreCase(XML_YES)) {
                        CEM cemItem = buscaComp (elementoCEM);                // Se obtienen todos los componentes asociados al CEM
                        cemX[n] = cemItem;
                        n++;
                    }
                }
            }

            fileDesc = fileDescX;
            cemNumber = cemNumberX;
            cem = cemX;
            comp = compX;
            cand.clear(); // Hay que eliminar los resultados anteriores del cálculo de candidatos.
        } catch(Exception e) {
            System.out.println(" *** Exception while loading XML file: " + e);
            e.printStackTrace();
        }
    }

    // Busca los componentes/ecuaciones asociados a un CEM y construye el objeto 
    // correspondiente con los posibles candidatos a fallo entre los que hay que buscar.
    public CEM buscaComp(Element elementoCEM) {
        CEM cemItem = new CEM();
        cemItem.name = elementoCEM.getAttribute(XML_CEM_NAME);
        String eqs [] = elementoCEM.getAttribute(XML_CEM_EQUATIONS).split(" ");
        NodeList listaCompAux;
        
        // Se obtiene el componente asociado a cada ecuación del CEM.
        for (int j = 0; j < eqs.length ; j++) {
            int encontrado = 0;
            String eqAux = eqs[j];
            String candidateName = "";
            String nomAux = "";
            listaCompAux = docXML.getElementsByTagName(XML_COMPONENT);
            for (int k = 0; k < listaCompAux.getLength(); k++) {
                Node nodo_k = listaCompAux.item(k);
                if ((nodo_k != null) && (nodo_k.getNodeType() == Node.ELEMENT_NODE)) {
                    Element elementoComponente = (Element) nodo_k;
                    NodeList listaHijos = elementoComponente.getChildNodes();
                    for (int l = 0; l < listaHijos.getLength(); l++) {
                        if (listaHijos.item(l).getNodeName().equals(XML_EQ)) {
                            nomAux = ((Element)listaHijos.item(l)).getAttribute(XML_ID);
                            if (eqAux.equals(nomAux)) {
                                if (this.tipoCandidatos.startsWith("Components")) {
                                    candidateName = elementoComponente.getAttribute(XML_COMP_NAME);
                                } else if (this.tipoCandidatos.startsWith("Equations")){
                                    candidateName = nomAux;  
                                }
                                encontrado = 1;
                                break;
                            }
                        }
                    }
                    if (encontrado == 1) {
                        break;
                    }
                } else if (nodo_k == null) {
                    System.out.println("There are errors in the XML structure.");
                    System.out.println("There is a null Component");
                }
            }
            if (encontrado == 1) {
                if (cemItem.candName.size() > 0) {
                    for (int p = 0; p < cemItem.candName.size(); p++){
                        if (cemItem.candName.get(p).equals(candidateName)) {
                            encontrado = 2;
                            break;
                        }
                    }
                }
            } else if (encontrado == 0){
                System.out.println("There are errors in the XML structure.");
                System.out.println("Couldn't find a Component with \"" + eqs[j] + "\";");
            }
            if (encontrado == 1) {
                cemItem.candName.add(candidateName);
                cemItem.candRef.add(cargaComp(candidateName));
            }
        }
        return cemItem;
    }

    public String getNombreCEM(int i) {
        return cem[i].name;
    }

    public CEM getCEM(int i) {
        return cem[i];
    }

    public void activarCEM(int i) {
        cem[i].considerFlag = true;
    }

    public void desactivarCEM(int i) {
        cem[i].considerFlag = false;
    }
}