/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.math.BigInteger;

/**
 *
 * @author José Ignacio Uribe
 */
// Generador de disposiciones con n1, n2,... nm elementos en cada una de las m posiciones.

public class DispositionGenerator {

    private int n[];                // Número de elementos en cada una de las m posiciones. Permanece fijo una vez ejecutado el constructor.
    private BigInteger total;       // Número total de disposiciones. Permanece fijo una vez ejecutado el constructor.
    private int[] v;                // Vector de disposiciones. Su tamaño permanece fijo una vez ejecutado el constructor.
    private BigInteger numLeft;     // Número de disposiciones todavía no generadas.

    // Constructor.
    public DispositionGenerator(int[] n) {
        if (n == null || n.length == 0) {
            throw new IllegalArgumentException();
        }
        for (int i = 0; i < n.length; i++) {
            if (n[i] < 1) {
                throw new IllegalArgumentException();
            }
        }
        this.n = n;
        
        total = new BigInteger(Integer.toString(n[0]));
        for (int i = 1; i < n.length; i++) {
            total = total.multiply(new BigInteger(Integer.toString(n[i])));
        }
        v = new int[n.length];
        
        reset();
    }

    // Repone el sistema de cálculo de disposiciones a la situación inicial.
    public void reset() {
        for (int i = 0; i < v.length; i++) {
            v[i] = 0;
        }
        numLeft = new BigInteger(total.toString()); // Quedan todas las disposiciones.
    }

    // Devuelve el número de disposiciones todavía no generadas.
    public BigInteger getNumLeft() {
        return numLeft;
    }

    // Devuelve un valor booleano que indica si quedan más disposiciones por generar.
    public boolean hasMore() {
        return (numLeft.compareTo(BigInteger.ZERO) == 1);
    }

    // Devuelve el número de disposiciones con n1, n2,... nm elementos en cada una de las m posiciones.
    public BigInteger getTotal() {
        return total;
    }

    // Devuelve la siguiente disposición.
    public int[] getNext() {
        if (numLeft.equals(total)) {
            numLeft = numLeft.subtract(BigInteger.ONE);
            return v;
        }
        for (int i = 0; i < v.length; i++) {
            if (++v[i] < n[i]) {
                break;
            }
            v[i] = 0;
        }
        numLeft = numLeft.subtract(BigInteger.ONE);
        return v;
    }
}
