/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Random;
import javax.swing.table.DefaultTableModel;
import matlabcontrol.MatlabConnectionException;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;
import matlabcontrol.MatlabProxyFactory;
import matlabcontrol.MatlabProxyFactoryOptions;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartFrame;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
//Módulo que realiza las tareas de generación de experimentos
public class GeneradorExperimentos {
    Boolean ficherosCargados;   //Indica si ya se han cargado los ficheros necesarios
    Boolean experimentoGenerado;   //indica si ya se han generado los modelos de simulación
    ModeloSistema modelo;       //Modelo de sistema con el que se trabaja
    DefaultTableModel modeloTabla;      //modelo de la tabla que almacena los resultados de un experimento
    ArrayList<Variable> entradas = new ArrayList<Variable>();   //Lista de variables de entrada con sus valores para cada instante
    ArrayList<Variable> condicionesIniciales = new ArrayList<Variable>();   //Lista de variables con los valores que tienen al inicio
    String rutaFuncionInstante;    //Ubicación del fichero donde se almacenará la función "instante" para Matlab
    ArrayList<Fallo> fallos = new ArrayList<Fallo>();   //Lista de fallos paramétricos
    
    //Parámetros del experimento
    int Tsim;   //nº de instantes de simulación
    double deltaT;      //paso de integración de las ecuaciones integrales
    double umbral;      //umbral de activación de PCs en el cálculo de candidatos
    double precisionSensores = 0.01;    // Varianza de la función Random que mete ruido a los sonsores.

    //Método de creación
    public GeneradorExperimentos(ModeloSistema m) {
        modelo = m;
        ficherosCargados = false;
        experimentoGenerado = false;
        Tsim = -1;
        deltaT = -1;
        umbral = -1;
    }

    //Establece a verdadero el atributo "ficherosCargados"
    public void setFicherosCargados() {
        this.ficherosCargados = true;
    }

    //Establece a verdadero el atributo "experimentoGenerado"
    public void setExperimentoGenerado() {
        this.experimentoGenerado = true;
    }

    public void setModeloTabla(DefaultTableModel mt) {
        this.modeloTabla = mt;
    }

    //Establece el valor del parámetro "instants"
    public void setTsim(int i) {
        this.Tsim = i;
    }

    //Establece el valor del parámetro "deltaT"
    public void setDeltaT(double d) {
        this.deltaT = d;
    }

    //Establece el valor del parámetro "umbral"
    public void setUmbral(double u) {
        this.umbral = u;
    }

    //Añade una variable de entrada
    public void addEntrada(Variable e) {
        this.entradas.add(e);
    }

    //Añade un fallo paramétrico
    public void addFallo(Fallo f) {
        this.fallos.add(f);
    }
 
    //Construye físicamente una función de Matlab que calcula la evolución
    //del sistema a cada instante, y la almacena en un fichero .m
    //Además, devuelve una lista con las variables del experimento
    public ArrayList<String> generarFuncionInstante() {
        
        this.rutaFuncionInstante = (new File(System.getProperty("user.dir"))).getParentFile().getAbsolutePath() + "\\FuncionInstante";  //Usar en la versión entregable
        //this.rutaFuncionInstante = System.getProperty("user.dir") + "\\FuncionInstante";  //Usar dentro de NetBeans o en la versión instalable
        File fichero = new File(rutaFuncionInstante);
        fichero.setExecutable(true, false);
        fichero.setReadable(true, false);
        fichero.setWritable(true, false);
        String nomFichero = rutaFuncionInstante + "\\" + "instant.m";
        String linea;
        try {
            BufferedWriter bw = new BufferedWriter(new FileWriter(nomFichero));
            linea = "function xNew = instant(AntStateVar,Ent,DeltaT,Param )";
            bw.write(linea);
            bw.newLine();
            bw.newLine();
            bw.write("% System inputs:");
            bw.newLine();

            for (int i = 0; i < entradas.size(); i++) {
                linea = entradas.get(i).nombre + " = Ent(" + (i+1) + ") ;";
                bw.write(linea);
                bw.newLine();
            }
            bw.newLine();
            bw.write("% Previous instant predictions for the state variables:");
            bw.newLine();

            Iterator i = modelo.variablesEstado.iterator();
            int c = 1;
            while (i.hasNext()) {
                String var = (String) i.next();
                linea = var + "_ant = AntStateVar(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            bw.newLine();
            bw.write("% System Parameters");
            bw.newLine();

            i = modelo.parametros.iterator();
            c = 1;
            while (i.hasNext()) {
                Parametro p = (Parametro) i.next();
                linea = p.nombre + " = Param(" + c + ") ;";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            bw.newLine();
            bw.write("% Transitional model equations:");
            bw.newLine();

            ArrayList<String> variables = new ArrayList<String>();
            i = modelo.modeloTransicionalOrd.iterator();
            while (i.hasNext()) {
                linea = (String) i.next();
                String[] partes = linea.split(" =");
                variables.add(partes[0]);
                bw.write(linea);
                bw.newLine();
            }
            bw.newLine();
            bw.write("% Observational model equations:");
            bw.newLine();
            
            i = modelo.modeloObservacional.iterator();
            while (i.hasNext()) {
                linea = (String) i.next();
                String[] partes = linea.split(" =");
                variables.add(partes[0]);
                bw.write(linea);
                bw.newLine(); 
            }
            bw.newLine();
            bw.write("% Outputs:");
            bw.newLine();

            i = variables.iterator();
            c = 1;
            while (i.hasNext()) {
                String v = (String) i.next();
                linea = "xNew(" + c + ") = " + v + ";";
                bw.write(linea);
                bw.newLine();
                c++;
            }
            bw.newLine();
            
            linea = "end";
            bw.write(linea);
            bw.close();
            return variables;
        } catch (IOException io) {
            System.out.println("Error while generating instant.m: " + io.getMessage());
        }
        return null;
    }

    //Ejecuta repetidamente la función "instante" tras crear una conexión con Matlab.
    //Los resultados de cada ejecución se almacenan por líneas en el modelo de la tabla
    public void ejecutarExperimento(String[] variables) throws MatlabConnectionException, MatlabInvocationException {
        //Creación de la tabla
        this.modeloTabla = new DefaultTableModel();
        modeloTabla.addColumn("Time");
        int numFallos = this.fallos.size();
        int numEntradas = this.entradas.size();
        Iterator it = this.fallos.iterator();
        while (it.hasNext()) {
            Fallo f = (Fallo) it.next();
            modeloTabla.addColumn(modelo.parametros.get(f.numParam - 1).nombre);
        }
        for (int i = 0; i < numEntradas; i++) {
            modeloTabla.addColumn(entradas.get(i).nombre);
        }
        for (int i = 0; i < variables.length; i++) {
            modeloTabla.addColumn(variables[i]);
        }

        MatlabProxyFactoryOptions options = new MatlabProxyFactoryOptions.Builder().setHidden(true).build();
        MatlabProxyFactory factory = new MatlabProxyFactory(options);
        MatlabProxy proxy = factory.getProxy();

        proxy.eval("addpath('" + this.rutaFuncionInstante + "')");
        int numVarEstado = modelo.variablesEstado.size();
        int numParametros = modelo.parametros.size();
        double[] AntStateVar = new double[numVarEstado];
        for (int i = 0; i < numVarEstado; i++) {
            String varEstado = modelo.variablesEstado.get(i);
            AntStateVar[i] = getValorCondicionInicial(varEstado, 0);
        }
        double[] Ent = new double[numEntradas];
        for (int i = 0; i < numEntradas; i++) {
            Ent[i] = getValorCondicionInicial(entradas.get(i).nombre, 0);
        }
        double DeltaT = this.deltaT;
        double[] Param = new double[numParametros];
        for (int i = 0; i < numParametros; i++) {
            Param[i] = modelo.parametros.get(i).valor;
        }
        Object result = proxy.returningFeval("instant", 1, AntStateVar, Ent, DeltaT, Param)[0];
        double[] x = (double[]) result;        
        Object[] fila = new Object[modeloTabla.getColumnCount()];
        
        // Iniciamos el contador de tiempo al valor que nos indican las C. I.
        double instanteActual = roundFourDecimals(getValorCondicionInicial("Time", 0));
        fila[0] = instanteActual;
        int c = 1;
        it = this.fallos.iterator();
        while (it.hasNext()) {
            Fallo f = (Fallo) it.next();
            fila[c] = roundFourDecimals(Param[f.numParam - 1]);
            c++;
        }
        for (int i = 0; i < numEntradas; i++) {
            fila[numFallos + 1 + i] = roundFourDecimals(Ent[i]);
        }
        for (int i = 0; i < x.length; i++) {
            fila[numFallos + numEntradas + 1 + i] = roundFourDecimals(x[i]);
        }
        modeloTabla.addRow(fila);
        
        // Para los instantes siguientes al inicio del experimento:
        instanteActual = roundFourDecimals(instanteActual + deltaT);
        int entradaActual = 1;
        while (instanteActual <= Tsim) {
            it = this.fallos.iterator();
            while (it.hasNext()) {          // Añadimos los fallos a los parámetros cuando sea necesario
                Fallo f = (Fallo) it.next();
                if ((new BigDecimal(instanteActual).setScale(1, 6)).compareTo(BigDecimal.valueOf(f.failStep)) == 0) {
                    if (f.tipo.compareTo("MULT") == 0) {
                        Param[f.numParam - 1] = Param[f.numParam - 1] * f.errorMagnitude;
                    } else if (f.tipo.compareTo("ADD") == 0) {
                        Param[f.numParam - 1] = Param[f.numParam - 1] + f.errorMagnitude;
                    }
                }
            }
            for (int i = 0; i < numEntradas; i++) {
                Ent[i] = entradas.get(i).valores.get(entradaActual - 1);
            }
            for (int j = 0; j < numVarEstado; j++) {
                AntStateVar[j] = x[x.length - ((numVarEstado*2) - j)];
            }
            result = proxy.returningFeval("instant", 1, AntStateVar, Ent, DeltaT, Param)[0];
            x = (double[]) result;
            
            // Ponemos los resultados en la tabla
            fila = new Object[modeloTabla.getColumnCount()];
            fila[0] = roundFourDecimals(instanteActual);
            c = 1;
            it = this.fallos.iterator();
            while (it.hasNext()) {
                Fallo f = (Fallo) it.next();
                fila[c] = roundFourDecimals(Param[f.numParam - 1]);
                c++;
            }
            for (int i = 0; i < numEntradas; i++) {
                fila[numFallos + 1 + i] = roundFourDecimals(Ent[i]);
            }
            Random rand = new Random();
            for (int i = 0; i < x.length; i++) { // Añadimos ruido en los sensores.
                if (i < x.length - numVarEstado) {
                    fila[numFallos + numEntradas + 1 + i] = roundFourDecimals(x[i]);
                } else {
                    fila[numFallos + numEntradas + 1 + i] = roundFourDecimals(x[i] + (rand.nextGaussian() * precisionSensores));
                }
            }
            modeloTabla.addRow(fila);
            instanteActual = roundFourDecimals(instanteActual + deltaT);
            entradaActual = entradaActual + 1;
        }
        proxy.exit();
        proxy.disconnect();
    }

    //Muestra un gráfico con los resultados obtenidos en el experimento
    //para una variable o parámetro de fallo
    public void mostrarGrafico(String variable) {
        int columna = modeloTabla.findColumn(variable);
        //System.out.println("Column to plot: " + columna + ", i.e. " + variable);
        XYSeriesCollection dataset = new XYSeriesCollection();
        XYSeries series = new XYSeries(variable);
        for (int i = 0; i < modeloTabla.getRowCount(); i++) {
            series.add(new Double(modeloTabla.getValueAt(i, 0).toString()), new Double(modeloTabla.getValueAt(i, columna).toString()));
        }
        dataset.addSeries(series);
        JFreeChart chart = ChartFactory.createXYLineChart(
                "Chart for " + variable, // Titulo
                "Time", // Etiqueta de datos
                "Value", // Etiqueta de valores
                dataset, // Datos
                PlotOrientation.VERTICAL, // orientacion
                true, // Incluye leyenda
                true, // Incluye tooltips
                false // urls
                );
        ChartFrame frame = new ChartFrame("Chart", chart);
        frame.pack();
        frame.setLocationRelativeTo(null);
        frame.setVisible(true);
    }

    //Muestra un gráfico con los resultados obtenidos en el experimento
    //para un conjunto de variables o parámetros de fallo
    public void mostrarGrafico(ArrayList<String> variables) {
        XYSeriesCollection dataset = new XYSeriesCollection();
        for (int v = 0; v < variables.size(); v++) {
            int columna = modeloTabla.findColumn(variables.get(v));
            XYSeries series = new XYSeries(variables.get(v));
            for (int i = 0; i < modeloTabla.getRowCount(); i++) {
                series.add(new Double(modeloTabla.getValueAt(i, 0).toString()), new Double(modeloTabla.getValueAt(i, columna).toString()));
            }
            dataset.addSeries(series);
        }
        JFreeChart chart = ChartFactory.createXYLineChart(
                "Chart for selected variables", // Titulo
                "Time", // Etiqueta de datos
                "Value", // Etiqueta de valores
                dataset, // Datos
                PlotOrientation.VERTICAL, // orientacion
                true, // Incluye leyenda
                true, // Incluye tooltips
                false // urls
                );
        ChartFrame frame = new ChartFrame("Chart", chart);
        frame.pack();
        frame.setLocationRelativeTo(null);
        frame.setVisible(true);
    }

    //Proporciona el valor inicial de una variable en un
    //instante previo al del inicio del experimento
    public double getValorCondicionInicial(String variable, int instante) {
        Iterator it = this.condicionesIniciales.iterator();
        while (it.hasNext()) {
            Variable v = (Variable) it.next();
            if (v.nombre.compareTo(variable) == 0) {
                return v.valores.get(instante);
            }
        }
        return -1;
    }

    public boolean configExpCargada() {
        if (Tsim < 0 || deltaT < 0 || umbral < 0) {
            return false;
        } else {
            return true;
        }
    }

    //Indica si se ha cargado la información necesaria para ejecutar
    //el módulo de GE. GeneradorExperimentos sólo es responsable de parte de
    //esa información; del resto se encarga ModeloSistema
    public boolean modeloCargadoModuloGE() {
        if (condicionesIniciales.isEmpty()) {
            System.out.println("Initial conditions did not load properly");
        }
        if (entradas.isEmpty()) {
            System.out.println("Input values did not load properly");
        }
        if (!configExpCargada()) {
            System.out.println("Experiment Configuration did not load properly");
        }
        if (condicionesIniciales.isEmpty() || entradas.isEmpty() || !configExpCargada()) {
            return false;
        } else {
            return true;
        }
    }

    public boolean modeloCargadoModuloDX() {
        return !(condicionesIniciales.isEmpty() || modeloTabla == null || !configExpCargada());
    }

    //Redondea valores a cinco decimales.
    public double roundFourDecimals(double d) {
        double result = d * 10000;
        result = Math.round(result);
        result = result / 10000;
        return result;
    }

    public void mostrarEntradas() {
        System.out.println("Input variables values: ");
        Iterator i = this.entradas.iterator();
        while (i.hasNext()) {
            Variable e = (Variable) i.next();
            System.out.println(e.nombre);
            Iterator j = e.valores.iterator();
            while (j.hasNext()) {
                System.out.println(" " + j.next());
            }
        }
    }
}
