/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.util.Random;
import javax.swing.JOptionPane;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
//Se encarga de generar los valores que tomarán las variables de
//entrada durante un experimento. Para ello lanza distintos formularios y
//con los datos introducidos por el usuario genera los valores siguiendo
//algún tipo de función.
//Actualmente se dispone de tres funciones distintas: constante, sinusoidal y de sierra
public class GeneradorFicheroEntradas {
    Interfaz ventana;
    String variable;    //nombre de la variable de entrada para la que se generarán valores
    boolean valido;     //indica si el proceso ha acabado correctamente y los datos cargados son válidos
    String tipoFuncion; //función elegida para generar los valores
    
    //Parámetros generales
    int Tsim;
    boolean usarRuido;
    double mediaRuido = 0;
    double desviacionRuido = 1;
    //Parámetros de la función constante
    double valorConstante;
    //Parámetros de la función sinusoidal
    double media;
    double periodo;
    double amplitud;
    //Parámetros de la función de sierra
    double valorMedio;
    double factorPeriodo;
    double factorAmplitud;

    //Establece a verdadero el valor del atributo "valido"
    public void setValido() {
        this.valido = true;
    }

    //Establece el valor del parámetro "Tsim"
    public void setTsim(int t) {
        this.Tsim = t;
    }

    //Establece a verdadero el valor del atributo "usarRuido"
    public void setUsarRuido() {
        this.usarRuido = true;
    }

    //Establece el valor del parámetro "mediaRuido"
    public void setMediaRuido(double m) {
        this.mediaRuido = m;
    }

    //Establece el valor del parámetro "desviacionRuido"
    public void setDesviacionRuido(double d) {
        this.desviacionRuido = d;
    }

    //Establece el valor del parámetro "valorConstante"
    public void setValorConstante(double v) {
        this.valorConstante = v;
    }

    //Establece el valor del parámetro "media"
    public void setMedia(double m) {
        this.media = m;
    }

    //Establece el valor del parámetro "amplitud"
    public void setAmplitud(double a) {
        this.amplitud = a;
    }

    //Establece el valor del parámetro "periodo"
    public void setPeriodo(double p) {
        this.periodo = p;
    }

    //Establece el valor del parámetro "valorMedio"
    public void setValorMedio(double v) {
        this.valorMedio = v;
    }

    //Establece el valor del parámetro "factorAmplitud"
    public void setFactorAmplitud(double a) {
        this.factorAmplitud = a;
    }

    //Establece el valor del parámetro "factorPeriodo"
    public void setFactorPeriodo(double p) {
        this.factorPeriodo = p;
    }

    //Método de creación nº 1
    //Lanza un formulario para recoger los parámetros más básicos
    public GeneradorFicheroEntradas(Interfaz i) {
        this.ventana = i;
        FormularioBasico fb = new FormularioBasico(ventana, true, this);
    }

    //Método de creación nº 2
    //Permite al usuario seleccionar el tipo de función que desee,
    //recoge los parámetros relativos a la misma, y genera los
    //valores para la variable de entrada indicada
    public GeneradorFicheroEntradas(Interfaz i, String v, int t) {
        this.ventana = i;
        this.variable = v;
        this.Tsim = t;
        this.valido = false;
        this.usarRuido = false;
        Object selectedValue = panelSeleccionFuncion();
        if (selectedValue != null && selectedValue.toString().compareTo("Constant Value") == 0) {
            System.out.println("A constant value function will be used");
            this.tipoFuncion = "constante";
            FuncionConstante fc = new FuncionConstante(ventana, true, this);
        } else if (selectedValue != null && selectedValue.toString().compareTo("Sinusoidal Function") == 0) {
            System.out.println("A sinusoidal function will be used");
            this.tipoFuncion = "sinusoidal";
            FuncionSinusoidal fs = new FuncionSinusoidal(ventana, true, this);
        } else if (selectedValue != null && selectedValue.toString().compareTo("Sawtooth Function") == 0) {
            System.out.println("A sawtooth function will be used");
            this.tipoFuncion = "sierra";
            FuncionSierra fs = new FuncionSierra(ventana, true, this);
        }
    }

    //Lanza la generación de valores para la variable de entrada indicada
    public double[] generarValoresEntrada() {
        if (tipoFuncion.compareTo("constante") == 0) {
            return generarConstante();
        } else if (tipoFuncion.compareTo("sinusoidal") == 0) {
            return generarSinusoidal();
        } else if (tipoFuncion.compareTo("sierra") == 0) {
            return generarSierra();
        }
        return null;
    }

    //Genera los valores siguiendo una función constante
    public double[] generarConstante() {
        int tamaño = Tsim*10;
        double[] d = new double[tamaño];
        Random rand = new Random();
        for (int i = 0; i < tamaño; i++) {
            if (usarRuido) {
                d[i] = valorConstante + (rand.nextGaussian() * desviacionRuido + mediaRuido);
            } else {
                d[i] = roundFourDecimals(valorConstante);
            }
            System.out.println("*** Value: " + d[i]);
        }
        return d;
    }

    //Genera los valores siguiendo una función sinusoidal
    public double[] generarSinusoidal() {
        int tamaño = Tsim*10;
        double [] tMuestra = new double [tamaño];
        for (int i = 0; i < tamaño; i++) {
            tMuestra[i] = 0.1*(i+1);
        }
        double[] d = new double[tamaño];
        Random rand = new Random();
        for (int i = 0; i < tamaño; i++) {
            double valor = media + amplitud * Math.sin(((2 * Math.PI * tMuestra[i]) / periodo));
            if (usarRuido) {
                d[i] = valor + (rand.nextGaussian() * desviacionRuido + mediaRuido);
            } else {
                d[i] = roundFourDecimals(valor);
            }
            System.out.println("*** Value: " + d[i]);
        }
        return d;
    }

    //Genera los valores siguiendo una función de sierra
    public double[] generarSierra() {
        int tamaño = Tsim*10;
        double[] d = new double[tamaño];
        double a1 = factorAmplitud * -0.02627;
        double a2 = factorAmplitud * -0.04886;
        double a3 = factorAmplitud * -0.03595;
        double a4 = factorAmplitud * -0.01895;
        double b1 = factorAmplitud * 0.1635;
        double b2 = factorAmplitud * 0.05584;
        double b3 = factorAmplitud * 0.01315;
        double b4 = factorAmplitud * -0.005875;
        double w = (1 / factorPeriodo) * 0.05236;
        //double w = ((2 * Math.PI) / factorPeriodo);
        Random rand = new Random();
        for (int i = 0; i < tamaño; i++) {
            double valor = valorMedio + a1 * Math.cos(i * w) + b1 * Math.sin(i * w)
                    + a2 * Math.cos(2 * i * w) + b2 * Math.sin(2 * i * w)
                    + a3 * Math.cos(3 * i * w) + b3 * Math.sin(3 * i * w)
                    + a4 * Math.cos(4 * i * w) + b4 * Math.sin(4 * i * w);
            if (usarRuido) {
                d[i] = valor + (rand.nextGaussian() * desviacionRuido + mediaRuido);
            } else {
                d[i] = roundFourDecimals(valor);
            }
            System.out.println("*** Value: " + d[i]);
        }
        return d;
    }

    //Diálogo que permite al usuario seleccionar el tipo de función a usar
    private Object panelSeleccionFuncion() {
        Object[] posiblesValores = {"Constant Value", "Sinusoidal Function", "Sawtooth Function"};
        Object selectedValue = JOptionPane.showInputDialog(ventana,
                "Select the type of function to use for variable " + this.variable, "Warning",
                JOptionPane.INFORMATION_MESSAGE, null, posiblesValores, posiblesValores[0]);
        return selectedValue;
    }
    
    //Redondea valores a cinco decimales.
    public double roundFourDecimals(double d) {
        double result = d * 10000;
        result = Math.round(result);
        result = result / 10000;
        return result;
    }
}
