/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */

//Módulo que realiza las tareas de generación de modelos de simulación
public class GeneradorModelosSimulacion {
    Boolean ficherosCargados;   //Indica si ya se han cargado los ficheros necesarios
    Boolean modelosGenerados;   //indica si ya se han generado los modelos de simulación
    ModeloSistema modelo;       //Modelo de sistema con el que se trabaja
    
    //Método de creación
    public GeneradorModelosSimulacion(ModeloSistema m){
        modelo = m;
        ficherosCargados = false;
        modelosGenerados = false;
    }
    
    //Establece a verdadero el atributo "ficherosCargados"
    public void setFicherosCargados(){
        this.ficherosCargados = true;
    }
    
    //Establece a verdadero el atributo "modelosGenerados"
    public void setModelosGenerados(){
        this.modelosGenerados = true;
    }
    
    //Función principal del proceso de generar modelos de simulación, en el que se recorre el MEM
    //desde el N.D. y se marca cada nodo para preparar las ecuaciones del modelo de simulación

    public ArrayList<String> generarModelosSimulacion(ArrayList<MEM> modelosSimulacion){
        String auxEc, auxVar, partes [];
        ArrayList <String> lista = new ArrayList <String>();
        for(int i=0;i<modelosSimulacion.size();i++){
            ArrayList <String> alFinal = new ArrayList <String>();
            ArrayList <String> desechar = new ArrayList <String>();
            MEM m = (MEM)modelosSimulacion.get(i);
            
            /* Debemos modificar el formato del número de Modelo y de Cadena,
               añadiendo ceros a la izquierda para que los modelos de simulación
               se ordenen por número de modelo en lugar de alfabéticamente */
            String cadenaAux = m.nombre;
            String nuevoNombre = "";
            Pattern patron = Pattern.compile("Modelo (.+)\\(Cadena (.+)\\)");
            Matcher matcher = patron.matcher(cadenaAux);
            matcher.find();

            int numModelo = Integer.parseInt(matcher.group(1));
            int numCadena = Integer.parseInt(matcher.group(2));
            if (modelosSimulacion.size() < 100) {
                nuevoNombre = String.format("Modelo %02d(Cadena %02d)", numModelo, numCadena);
            } else {
                nuevoNombre = String.format("Modelo %03d(Cadena %03d)", numModelo, numCadena);
            }
            lista.add(nuevoNombre + "Simulacion.sim");
            
            // Si se trata de un modelo con dos formas de calcular la variable
            // discrepancia, debemos renombrar una de las variables cabeza esas
            // ecuaciones, para que la segunda no sobreescriba la primera
            for (int j=0; j<modelo.modeloObservacional.size(); j++) {
                int bandera = 0;
                auxEc = modelo.modeloObservacional.get(j);
                partes = auxEc.split( " = ");
                auxVar = partes[1].replace("_ant ;", "");
                for (int k = m.interpretaciones.size()-1; k>=0; k--) {
                    Interpretacion interp = m.interpretaciones.get(k);
                    if (interp.variableCabeza.equals(auxVar)) {
                        if (interp.nombre.split("_")[1].equals("2")) {
                            for (int l=0; l<m.interpretaciones.size();l++) {
                                Interpretacion interp2 = m.interpretaciones.get(l);
                                String auxVar2 = interp2.variableCabeza;
                                if (auxVar.equals(auxVar2) && k!=l) {
                                    auxEc = partes[1].replace(" ;", "") + " = " + partes[0];
                                    lista.add(auxEc.replaceAll("_ant", "_2"));
                                    bandera = 1;
                                    break;
                                }
                            }
                            if (bandera == 0) {
                                auxEc = auxVar + " = " + partes[0];
                                desechar.add(auxVar);
                                lista.add(auxEc);
                                bandera = 1;
                                break;
                            }
                        }
                    } else {
                        if (interp.variableCabeza.equals(partes[0])) {
                            bandera = 1;
                            alFinal.add(auxEc.replaceAll("_ant",""));
                        }
                    }
                    if (bandera == 1) {
                        break;
                    }
                }
            }
            
            for (int j=0; j<modelo.modeloTransicionalOrd.size();j++) {
                auxEc = modelo.modeloTransicionalOrd.get(j);
                String varCabeza = auxEc.split(" = ")[0];
                for (int k = 0; k < m.interpretaciones.size();k++) {
                    Interpretacion interp = m.interpretaciones.get(k);
                    if (varCabeza.equals(interp.variableCabeza) && !(desechar.contains(varCabeza))) {
                        for (int l=0; l<desechar.size(); l++) {
                            auxEc = auxEc.replace(desechar.get(l) + "_ant",desechar.get(l));
                        }
                        lista.add(auxEc);
                        break;
                    }
                }
            }
            // A continuación, añadimos las ecuaciones que habíamos dejado
            // para el final, del tipo Var_medida = Var_estado.
            for (int j=0; j<alFinal.size(); j++) {
                lista.add(alFinal.get(j));
            }
        }
        return lista;
    }
}