/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.util.ArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
public class LogicaDecision {
    Diagnosis moduloDX;
    ArrayList<ArrayList<String[]>> candidatos = new ArrayList<ArrayList<String[]>>();
    ArrayList<String> modelosActivados;
    ArrayList<Double> instantesActivacion = new ArrayList<Double>();
    int VRsize;
    int VMsize;
    int desfase;
    double alpha;

    public LogicaDecision(Diagnosis m) {
        this.moduloDX = m;
        this.modelosActivados = new ArrayList<String>();
    }

    public void setVRsize(int t) {
        this.VRsize = t;
    }

    public void setVMsize(int t) {
        this.VMsize = t;
    }

    public void setDesfase(int d) {
        this.desfase = d;
    }

    public void setAlpha(double a) {
        this.alpha = a;
    }

    public void calcularCandidatos(ModeloSistema modelo) {
        int numDiscrepancias = moduloDX.nodosDiscrepancia.size();
        boolean[] banderas = new boolean[numDiscrepancias];
        for (int i = 0; i < banderas.length; i++) {
            banderas[i] = false;
        }

        int posicionInicio = 0;     //podría pedírsele al usuario en FormularioZtest
        int posicionFinal = posicionInicio + VRsize + VMsize + desfase - 1;
        String nodoDiscrepancia;
        double[] ventanaReferencia;
        double[] ventanaMuestra;
        double mediaVR, sigmaVR;
        double zVal;
        double significancia = 1;
        Ztest test;
        int numInstantes = moduloDX.modeloTabla.getRowCount();
        while (posicionFinal < numInstantes) {
            for (int j = 0; j < numDiscrepancias; j++) {
                if (!banderas[j]) {     //si un modelo se activa, permanece así hasta el final y ya no se mira si tiene más activaciones
                    String nombreModelo = modelo.listaModSim.get(j).nombreModelo;
                    nodoDiscrepancia = moduloDX.nodosDiscrepancia.get(j);
                    ventanaReferencia = obtenerVR(posicionInicio, nombreModelo);
                    ventanaMuestra = obtenerVM(posicionInicio + VRsize + desfase, nombreModelo);
                    mediaVR = obtenerMediaVentana(ventanaReferencia);
                    sigmaVR = obtenerSigmaVentana(ventanaReferencia, mediaVR);
                    test = new Ztest(ventanaMuestra, mediaVR, sigmaVR);
                    zVal = test.getZVal();
                    
                    if (!Double.isNaN(zVal)) {
                        significancia = jsc.distributions.Normal.standardTailProb(zVal, true);
                    }
                    if (significancia < alpha) {
                        Pattern patron = Pattern.compile("Modelo (\\d+)\\(Cadena (\\d+)\\)");
                        Matcher matcher = patron.matcher(nombreModelo);
                        matcher.find();
                        int cadena = Integer.parseInt(matcher.group(2));
                        String cadenaAux;
                        int index = 0;
                        for (int k = 0; k < moduloDX.cemNumber; k++) {
                            cadenaAux = moduloDX.cem[k].name;
                            if (cadenaAux.contains(" " + cadena)) {
                                index = k;
                                break;
                            }
                        }
                        System.out.print("\nVRef: ");
                        for (int i = 0; i < VRsize; i++) {
                            System.out.print(" " + ventanaReferencia[i]);
                        }
                        System.out.print("\nVMuestra: ");
                        for (int i = 0; i < VMsize; i++) {
                            System.out.print(" " + ventanaMuestra[i]);
                        }
                        System.out.println("\nMedia: " + mediaVR + ", sigma: " + sigmaVR);
                        System.out.println("\nSignificance: " + significancia + ", alpha: " + alpha + "; zVal " + zVal);

                        banderas[j] = true;
                        moduloDX.activarCEM(index);     //ATENCION: se presupone que el orden de los MEMs y CEMs es el mismo
                        modelosActivados.add(nombreModelo);
                        instantesActivacion.add(moduloDX.getValorTablaRes(posicionFinal, "Time"));
                        moduloDX.calculaCandidatos(modelo);
                        candidatos.add((ArrayList<String[]>) moduloDX.cand.clone());
                        System.out.println("*** Node " + nodoDiscrepancia + " activated at " + moduloDX.getValorTablaRes(posicionFinal, "Time") + " seconds.");
                    }
                }
            }
            posicionInicio++;
            posicionFinal++;
        }
    }

    public double[] obtenerVR(int inicio, String nd) {
        double[] ventana = new double[VRsize];
        for (int i = 0; i < ventana.length; i++) {
            ventana[i] = moduloDX.getValorResiduoTablaRes(i + inicio, nd);
        }
        return ventana;
    }

    public double[] obtenerVM(int inicio, String nd) {
        double[] ventana = new double[VMsize];
        for (int i = 0; i < ventana.length; i++) {
            ventana[i] = moduloDX.getValorResiduoTablaRes(i + inicio, nd);
        }
        return ventana;
    }

    public double obtenerMediaVentana(double[] ventana) {
        double sum = 0;
        for (int i = 0; i < ventana.length; i++) {
            sum = sum + ventana[i];
        }
        return sum / ventana.length;
    }

    public double obtenerSigmaVentana(double[] ventana, double media) {
        double sum = 0;
        for (int i = 0; i < ventana.length; i++) {
            sum = sum + Math.pow((ventana[i] - media), 2);
        }
        return Math.sqrt(sum / (ventana.length - 1));
    }
}