/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.util.ArrayList;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
public class ModeloSimulacion {
    String nombreModelo;
    String nombreFichero;
    ArrayList<Variable> nodoDiscrepancia;
    ModeloSistema modeloSistema;
    ArrayList<String> listaEcuaciones;
    ArrayList<Interpretacion> interpretaciones;
    ArrayList<String> variablesEstado = new ArrayList<String>();    //Lista de variables de estado
    ArrayList<String> variablesSalida = new ArrayList<String>();    //Lista de variables de salida
    ArrayList<String> variablesEntrada = new ArrayList<String>();    //Lista de variables de entrada
    
    public ModeloSimulacion(String nombre, String fichero, ModeloSistema modeloSist){
        nombreModelo = nombre;
        nombreFichero = fichero;
        modeloSistema = modeloSist;
        nodoDiscrepancia = new ArrayList();
        listaEcuaciones = new ArrayList();
        interpretaciones = new ArrayList<Interpretacion>();
        variablesEstado = new ArrayList<String>();
        variablesSalida = new ArrayList<String>();
        variablesEntrada = new ArrayList<String>();
    }
    
    public void addEcuacion(String e){
        listaEcuaciones.add(e);
    }
    
    // Esta función calcula el nodo discrepancia del modelo de un posible
    // conflicto, es decir, el nodo sobre el que se calcula el residuo en cada
    // caso. Existen dos posibilides: O bien se calcula un nodo que a la vez es
    // medido, o se calcula una variable por dos caminos.
    public void findNodoDiscrepancia(){
        for (int i=0;i<listaEcuaciones.size();i++) {
            String ec = (String)listaEcuaciones.get(i);
            String var = ec.split(" = ")[0];
            if (modeloSistema.variablesSalida.contains(var)) {  // Comprobamos si alguna variable medida es cabeza de una de las ecuaciones del modelo del P. C.
                nodoDiscrepancia.add(new Variable(var));
                return;
            } else {                                            // Si no, comprobamos si alguna variable se calcula dos veces en el posible conflicto.
                String var2 = var + "_2";
                for (int j=0;j<listaEcuaciones.size();j++) { 
                    if (j!=i) {
                        String ecAux = (String)listaEcuaciones.get(j);
                        String varAux = ecAux.split(" = ")[0];
                        if (var2.equals(varAux)) {
                            nodoDiscrepancia.add(new Variable(var));
                            nodoDiscrepancia.add(new Variable(var2));
                            return;
                        }
                    }
                }
            }
        }   
    }
    
    // La siguiente función actualiza las siguientes propiedades del modelo de simulación:
    //    - variablesSalida
    //    - variablesEntrada
    //    - variablesEstado
    public void initVariables(){
        ArrayList <String> varAnt = new ArrayList <String>();
        for (int j=0; j<modeloSistema.variablesEstado.size(); j++) {
            varAnt.add(modeloSistema.variablesEstado.get(j) + "_ant");
        }
        for (int i=0;i<listaEcuaciones.size();i++) { // Comprobamos las variables de cada tipo que hay en cada ecuación y actualizamos el correspondiente ArrayList
            String ec = (String)listaEcuaciones.get(i);
            String elements [] = ec.split(" ");
            for (int j=0; j<elements.length; j++) {
                if (modeloSistema.variablesSalida.contains(elements[j]) && !this.variablesSalida.contains(elements[j])) {
                    this.variablesSalida.add(elements[j]);
                } else if (modeloSistema.variablesEntrada.contains(elements[j]) && !this.variablesEntrada.contains(elements[j])){
                    this.variablesEntrada.add(elements[j]);
                } else if (varAnt.contains(elements[j]) && !this.variablesEstado.contains(elements[j].replace("_ant", ""))){
                    this.variablesEstado.add(elements[j].replace("_ant", ""));
                }
            }
        }
    }
    
    public void mostrar(){
        System.out.println("\n"+this.nombreModelo);
        for(int i=0; i<listaEcuaciones.size(); i++){
            System.out.println("    " + listaEcuaciones.get(i));
        }
    }   
}
