/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package generadorexperimentosdiagnosis;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Stack;

/**
 *
 * @author David Rubio + Luis Miguel Villarroel
 */
//Clase que almacena y gestiona la información que introduce
//el usuario en el sistema.
public class ModeloSistema {

    String nombre;
    ArrayList<String> listaEcuaciones = new ArrayList<String>();    //Lista de ecuaciones en formato texto
    ArrayList<String> modeloObservacional = new ArrayList<String>();    //Lista de ecuaciones que forman el modelo observacional
    ArrayList<String> modeloTransicional = new ArrayList<String>();    //Lista de ecuaciones que forman el modelo transicional
    ArrayList<String> modeloTransicionalOrd = new ArrayList<String>();    //Lista ordenada de las ecuaciones del modelo transicional
    ArrayList<String> listaEcuacionesIntegrales = new ArrayList<String>();  //Lista de ecuaciones integrales en formato texto
    ArrayList<String> listaVariables = new ArrayList<String>();  //Lista de variables en formato texto
    ArrayList<String> modeloCPC = new ArrayList<String>();      //Contenido de un CMEM en formato texto
    ArrayList<MEM> modelos = new ArrayList<MEM>();      //Lista de MEMs con estructura definida
    ArrayList<String> variablesEstado = new ArrayList<String>();    //Lista de variables de estado
    ArrayList<String> variablesSalida = new ArrayList<String>();    //Lista de variables de salida
    ArrayList<String> variablesEntrada = new ArrayList<String>();    //Lista de variables de entrada
    ArrayList<String> variablesUsadas = new ArrayList<String>();    //Lista de variables usadas en el algoritmo de recorrido del árbol causal
    ArrayList<String> listaModelosSimulacion = new ArrayList<String>();     //Lista de modelos de simulación GENERADOS (una línea de texto por ecuación)
    //ArrayList<String> nodosDiscrepancia = new ArrayList<String>(); // Lista de todos los nodos discrepancia del sistema.
    ArrayList<Parametro> parametros = new ArrayList<Parametro>();   //Lista con los parámetros
    ArrayList<String> modeloES;     //Modelo de un sistema en notación Entrada / Salida, almacenado por líneas.
    //ArrayList<String> listaModSim = new ArrayList<String>();     //Lista de modelos de simulación CARGADOS  <--- meter en DOC.
    ArrayList<ModeloSimulacion> listaModSim = new ArrayList<ModeloSimulacion>();

    final String [] simbolos = { ";", "(", ")", ".*", "./", "+", "-", ".^"};
    
    //Añade una ecuación en formato texto a la lista
    public void addEcuacionLista(String s) {
        listaEcuaciones.add(s);
    }
    
    public void addEcModObs(String s) {
        modeloObservacional.add(s);
    }
    
    public void addEcModTrans(String s) {
        modeloTransicional.add(s);
    }

    //Añade una ecuación integral en formato texto a la lista
    public void addEcuacionIntegralLista(String s) {
        listaEcuacionesIntegrales.add(s);
    }

    //Añade una línea con variables en formato texto a la lista
    public void addVarEntradaLista(String s) {
        listaVariables.add(s);
    }

    //Añade una línea de un CMEM en formato texto a la lista
    public void addLineaCPC(String s) {
        modeloCPC.add(s);
    }

    //Añade un MEM a la lista
    public void addMEM(MEM m) {
        modelos.add(m);
    }

    //Añade una variable de estado a la lista
    public void addVariableEstado(String s) {
        variablesEstado.add(s);
    }

    //Añade una variable de salida a la lista
    public void addVariableSalida(String s) {
        variablesSalida.add(s);
    }

    //Añade una variable de entrada a la lista
    public void addVariableEntrada(String s) {
        variablesEntrada.add(s);
    }

    //Añade un parámetro a la lista
    public void addParametro(Parametro p) {
        parametros.add(p);
    }
    
    public void addModSim(ModeloSimulacion ms) {
        listaModSim.add(ms);
    }

    //Establece la lista que contiene un modelo de
    //sistema en notación Entrada / Salida
    public void setModeloES(ArrayList<String> mod) {
        modeloES = mod;
    }

    //Dice si se ha cargado la información necesaria para ejecutar
    //el módulo de CM
    public boolean modeloCargadoModuloCM() {
        return !(listaEcuaciones.isEmpty() || listaVariables.isEmpty());
    }

    //Indica si se ha cargado la información necesaria para ejecutar
    //el módulo de GE. ModeloSistema sólo es responsable de parte de
    //esa información; del resto se encarga GeneradorExperimentos
    public boolean modeloCargadoModuloGE() {
        return !(listaEcuaciones.isEmpty() || listaVariables.isEmpty() || parametros.isEmpty());
    }

    //Dice si se ha cargado la información necesaria para ejecutar
    //el módulo de GMS
    public boolean modeloCargadoModuloGMS() {
        return !(listaEcuaciones.isEmpty() || listaVariables.isEmpty() || modeloCPC.isEmpty());
    }

    public boolean modeloCargadoModuloDX() {
        if (!listaVariables.isEmpty()) {
            if (!modeloCPC.isEmpty()) {
                if (!listaModSim.isEmpty()) {
                    return true;
                } else {
                    System.out.println("Simulation Models List Failure");
                    return false;
                }
            } else {
                System.out.println("CPC Model Failure");
                return false;
            }
        } else {
            System.out.println("Entry Variables Failure");
            return false;
        }
    }

    //Dice si se ha cargado la información necesaria para ejecutar
    //el módulo de generación de entradas
    public boolean modeloCargadoGeneradorEntradas() {
        return !variablesEntrada.isEmpty();
    }

    //Dice si se ha cargado la información necesaria para ejecutar
    //el módulo de generación de fallos paramétricos
    public boolean modeloCargadoGeneradorFallos() {
        return !parametros.isEmpty();
    }
    
    // Construye los modelos observacional y transicional a partir del sistema
    // cargado. El objetivo es buscar y ordenar las ecuaciones con las que se
    // generarán las funciones de simulación.
    public void prepararSistema() {
        Iterator i = listaEcuaciones.iterator();
        String ec;
        ArrayList <String> varCandidatas = new ArrayList<String>();
        while (i.hasNext()) {
            ec = (String) i.next();
            String[] partes = ec.split(" = ");
            String cabeza = partes[0];
            partes[1] = partes[1].replace(" ;","");
            ArrayList <String> varCola = new ArrayList <String>();
            varCola.toArray(partes[1].split(" "));
            if (variablesSalida.contains(cabeza)) {                      // Buscamos las ecuaciones del modelo observacional
                // Comprobamos si dependen directamente de una variable de estado
                // Si es así, debemos cambiar ésta por una que refleje el valor
                // en el instante anterior. Búsqueda recursiva sobre las variables
                // de cola de la ecuación.
                String var = comprobarEcModObs(ec);
            } else if (!modeloObservacional.contains(ec)){               // A continuación las ecuaciones del modelo transicional.
                addEcModTrans(ec);
            } 
            for (int k=0;k<varCola.size();k++) {
                if (!varCandidatas.contains(varCola.get(k))) {
                    varCandidatas.add(varCola.get(k));
                }
            }
        }
        
        ordenarModTrans();
        
        i = listaEcuacionesIntegrales.iterator();
        while (i.hasNext()) {                                                   // Después todas las ecuaciones integrales que hemos creado
            ec = (String) i.next();                                             // durante la carga de variables para modelar el paso del 
            modeloTransicionalOrd.add(ec);                                      // tiempo en las variables de estado.
        }
        
    }
    
    // Función de apoyo para comprobar las ecuaciones que debemos incluir en el
    // modelo observacional
    private String comprobarEcModObs(String ec) {
        String auxEc, auxVar, var = "";
        String[] partes = ec.split(" = ");
        partes[1] = partes[1].replace(" ;","");
        String [] varCola = partes[1].split(" ");
        int i = 0;
        
        while(i<varCola.length) {    
            auxVar = varCola[i];
            auxEc = buscaEcuacion(auxVar);
            if (listaEcuacionesIntegrales.contains(auxEc)) {
                ec = ec.replaceAll(auxVar + " ",auxVar + "_ant ");
                addEcModObs(ec);
                return auxVar;
            } else if (!auxEc.equals("")) {
                var = comprobarEcModObs(auxEc);
            }
            i++;
        }
        if (variablesEstado.contains(var)) {
            ec = ec.replaceAll(var + " ",var + "_ant ");
        }
        addEcModObs(ec);
        return "";
    }    
    
    /*
    // Ordena las ecuaciones del modelo transicional, de manera que éstas se
    // presenten en orden causal.
    public void ordenarModTrans() {
        Iterator i = variablesEstado.iterator();
        String ec;
        ArrayList <String> varCandidatas = new ArrayList<String>();
        Stack <String> pila = new Stack <String>();
        Stack <String> pilaAux = new Stack <String>();
        
        // Buscamos las raíces del árbol causal
        while (i.hasNext()) {
            ec = buscaEcuacion((String) i.next());
            String[] partes = ec.split(" = ");
            String cabeza = partes[0];
            if (variablesEstado.contains(cabeza)) {
                pilaAux.push(ec);
            }
            String [] varCola = partes[1].split(" ");
            for (int j=0;j<varCola.length;j++) {
                if (!varCandidatas.contains(varCola[j])) {
                    varCandidatas.add(varCola[j]);
                }
            }
        }
        
        // Ordenamos las ecuaciones de las raíces tal como las leemos, en orden
        // inverso a como están en la pila auxiliar.
        int tamañoPila = pilaAux.size();
        for(int k=0;k<tamañoPila;k++) {
            pila.push(pilaAux.pop());
        }
        
        // Una vez tenemos las raíces, debemos descender por el árbol causal
        // con una función recursiva que analice cada rama.
        while(!varCandidatas.isEmpty()) {
            nextTreeStep(varCandidatas, pila);
        }
        
        // Desapilamos las ecuaciones que haya en la pila hacia el modelo transicional
        tamañoPila = pila.size();
        for(int k=0;k<tamañoPila;k++) {
            modeloTransicionalOrd.add(pila.pop());
        }
    }
    
    // Esta función recursiva auxiliar es la que realmente realiza el proceso
    // de recorrer el árbol causal para reordenar las ecuaciones
    public void nextTreeStep(ArrayList <String> variables, Stack <String> pila) {
        int contador = 0;            // posición en la que insertaremos las nuevas variables de cola
        if (variables.isEmpty()) {
            return;
        } else {
            String nodo = variables.get(0);
            variables.remove(0);
            
            if (nodo.equals(";")) {
                //Desapilamos la primera ecuación de la pila
                String ec = pila.pop();
                if (!modeloTransicionalOrd.contains(ec)) {
                    modeloTransicionalOrd.add(ec);
                }
            }else if (!variablesUsadas.contains(nodo)) {
                variablesUsadas.add(nodo);
                String aux = buscaEcuacion(nodo);
                if (!aux.equals("") && !pila.contains(aux)) {
                    pila.push(aux);
                    String [] varCola = aux.split(" ");
                    for (int i=2;i<varCola.length;i++) {
                        if (!variables.contains(varCola[i])) {
                            variables.add(contador, varCola[i]);
                            contador ++;
                        }
                    }
                } else if (variablesEntrada.contains(nodo)) {
                    // Hemos llegado a una hoja que es variable de entrada;
                    return;
                } else if (variablesEstado.contains(nodo)) {
                    // Llegamos a un nodo hoja. Debemos cambiar el nodo por nodo(t-1)
                    aux = pila.pop();
                    aux = aux.replaceAll(nodo + " ", nodo + "_ant ");
                    pila.push(aux);
                    // Sacamos la variable de estado de la lista de usadas, pues
                    // debemos cambiarla en todas las ecuaciones en las que 
                    // aparezca.
                    variablesUsadas.remove(nodo);
                    return;
                } else {
                    for (int i = 0; i < parametros.size();i++) {
                        if (parametros.get(i).nombre.equals(nodo)) {
                            return;
                        }
                    }
                }
            }
        }
        nextTreeStep(variables, pila);
    }
    */
    
    // Ordena las ecuaciones del modelo transicional, de manera que éstas se
    // presenten en orden causal.
    public void ordenarModTrans() {
        String ec, ecAux, varAux;
        ArrayList <String> ecFinales = new ArrayList<String>();
        Stack <String> pila = new Stack <String>();
        
        // Comprobamos si tenemos lo que hace falta para calcular cada ecuación
        for (int i = 0; i < modeloTransicional.size(); i++) {
            int bandera = 0;
            ec = (String) modeloTransicional.get(i);
            String[] partes = ec.split(" = ");
            String cabeza = partes[0];
            String [] varCola = partes[1].split(" ");
            for (int j=0;j<varCola.length;j++) {
                varAux = varCola[j];
                int k = 0;
                for (k = 0; k < modeloTransicional.size(); k++) { 
                    if (i <= k) {
                        continue;
                    }
                    ecAux = (String) modeloTransicional.get(k);
                    String[] partesAux = ecAux.split(" = ");
                    String cabezaAux = partesAux[0];
                    if (cabezaAux.equals(varAux)) {
                        break;
                    }
                }
                
                if (k < modeloTransicional.size()) {               //Ecuación mal posicionada.
                    ecFinales.add(ec);
                    bandera = 1;
                    break;
                } else if (variablesEstado.contains(varAux)){
                    ec = ec.replaceAll(varAux + " ", varAux + "_ant ");
                }
            }
            if (bandera == 0) {
                modeloTransicionalOrd.add(ec);
            }
        }
        if (ecFinales.size() > 0) {
            modeloTransicional = (ArrayList <String>) ecFinales.clone();
            ordenarModTrans();
        }
    }
    
    //Función de soporte que devuelve aquella ecuación del sistema
    //que se corresponda con la variable de cabeza de una interpretación de un MEM
    public String buscaEcuacion(String cabeza){
        String aux, cadena = "";
        for(int i=0;i<listaEcuacionesIntegrales.size();i++){
            aux = listaEcuacionesIntegrales.get(i);
            if(aux.startsWith(cabeza+" ")){
                cadena = aux;
            }
        }
        if (cadena.equals("")) {
            for(int i=0;i<listaEcuaciones.size();i++){
                aux = listaEcuaciones.get(i);
                if(aux.startsWith(cabeza+" ")){
                    cadena = aux;
                }
            }
        }
        return cadena;
    }
}
